# COPYRIGHT 2018 ESRI
#
# TRADE SECRETS: ESRI PROPRIETARY AND CONFIDENTIAL
# Unpublished material - all rights reserved under the
# Copyright Laws of the United States.
#
# For additional information, contact:
# Environmental Systems Research Institute, Inc.
# Attn: Contracts Dept
# 380 New York Street
# Redlands, California, USA 92373
#
# email: contracts@esri.com

from arcpy.ia import Apply

__all__ = [
    'Apply',
    'ArgStatistics',
    'Arithmetic',
    'Aspect',
    'AspectSlope',
    'BandArithmetic',
    'CIg',
    'CIre',
    'ClayMinerals',
    'Colormap',
    'ColormapToRGB',
    'ColorspaceConversion',
    'Complex',
    'CompositeBand',
    'Contour',
    'ContrastBrightness',
    'Convolution',
    'Curvature',
    'EVI',
    'ElevationVoidFill',
    'ExtractBand',
    'FerrousMinerals',
    'GEMI',
    'GNDVI',
    'GVITM',
    'Grayscale',
    'HeatIndex',
    'Hillshade',
    'IronOxide',
    'Lookup',
    'MLClassify',
    'MSAVI',
    'MTVI2',
    'Mask',
    'NDVI',
    'NDVIre',
    'NDWI',
    'PVI',
    'Pansharpen',
    'RTVICore',
    'RasterCalculator',
    'Remap',
    'Resample',
    'SAVI',
    'SRre',
    'ShadedRelief',
    'Slope',
    'Speckle',
    'SpectralConversion',
    'Statistics',
    'StatisticsHistogram',
    'Stretch',
    'Sultan',
    'TSAVI',
    'TasseledCap',
    'Threshold',
    'TransposeBits',
    'UnitConversion',
    'VARI',
    'VectorField',
    'VectorFieldRenderer',
    'WindChill',
]


def Grayscale(raster, conversion_parameters=None):
    """
    The Grayscale function converts a multi-band image into a single-band Grayscale
    image. Specified weights are applied to each of the input bands, and a
    normalization is applied for output.

    Parameters:
      raster(Raster): the input raster
      conversion_parameters(List): array of double (A length of N array representing weights for each band, where N=band count.)

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    template_dict = {
        "rasterFunction": "Grayscale",
        "rasterFunctionArguments": {}
    }

    if conversion_parameters is not None and isinstance(conversion_parameters, list):
        template_dict["rasterFunctionArguments"]['ConversionParameters'] = conversion_parameters
    return Apply(in_raster=raster, raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def Arithmetic(raster1, raster2, operation_type="Plus", extent_type="FirstOf", cellsize_type="FirstOf"):
    """
    The Arithmetic function performs an Arithmetic operation between two rasters or a raster and a scalar, and vice versa.

    Parameters:
      raster1(Raster): the first raster (can be scalar)
      raster2(Raster): the 2nd raster (can be scalar)
      operation_type(string): one of "Plus", "Minus", "Multiply", "Divide", "Power", "Mode"
      extent_type(string): one of "FirstOf", "IntersectionOf" "UnionOf", "LastOf"
      cellsize_type(string): one of "FirstOf", "MinOf", "MaxOf "MeanOf", "LastOf"

    Returns:
      output_raster(Raster): The output raster with the function applied to it
    """
    extent_types = {
        "FirstOf": 0,
        "IntersectionOf": 1,
        "UnionOf": 2,
        "LastOf": 3
    }

    cellsize_types = {
        "FirstOf": 0,
        "MinOf": 1,
        "MaxOf": 2,
        "MeanOf": 3,
        "LastOf": 4
    }

    operation_types = {
        "Plus": 1,
        "Minus": 2,
        "Multiply": 3,
        "Divide": 4,
        "Power": 5,
        "Mode": 6
    }

    if extent_type not in extent_types.keys():
        raise ValueError(
            "Invalid extent_type value. Note that operation type should be one fo 'FirstOf', 'IntersectionOf', 'UnionOf', 'LastOf'")
    in_extent_type = extent_types[extent_type]

    if cellsize_type not in cellsize_types.keys():
        raise ValueError(
            "Invalid extent_type value. Note that operation type should be one fo 'FirstOf', 'MinOf', 'MaxOf', 'MeanOf', 'LastOf'")
    in_cellsize_type = cellsize_types[cellsize_type]

    if operation_type not in operation_types.keys():
        raise ValueError(
            "Invalid operation_type value. Note that operation type should be one fo 'Plus', 'Minus', 'Multiply', 'Divide', 'Power', 'Mode'")
    in_operation_type = operation_types[operation_type]

    template_dict = {
        "rasterFunction": "Arithmetic",
        "rasterFunctionArguments": {
            "Operation": in_operation_type
        }
    }

    if in_extent_type is not None:
        template_dict["rasterFunctionArguments"]['ExtentType'] = in_extent_type
    if in_cellsize_type is not None:
        template_dict["rasterFunctionArguments"]['CellsizeType'] = in_cellsize_type
    return Apply(in_raster=(raster1, raster2),
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def ArgStatistics(rasters, stat_type=None, min_value=None, max_value=None, multiple_occurrence_value=100):
    """
    The ArgStatistics function produces an output with a pixel value that represents a statistical metric from all
    bands of input rasters. The statistics can be the band index (or the slice index for multidimensional raster) of the maximum, minimum, or median value, or the
    duration (number of bands) between a minimum and maximum value

    Parameters:
      rasters(list of Rasters): Input rasters
      stat_type(string): one of "max", "min", "median", "duration"
      min_value(double): required if the type is duration
      max_value(double): required if the type is duration
      multiple_occurrence_value(int): the value will be returned when the max or min value occurred at multiple bands or slices. required if stat_ype is max or min

    Returns:
      output_raster(Raster): The output raster
    """
    stat_types = {
        'max': 0,
        'min': 1,
        'median': 2,
        'duration': 3
    }

    template_dict = {
        "rasterFunction": "ArgStatistics",
        "rasterFunctionArguments": {}
    }

    if stat_type.lower() not in stat_types.keys():
        raise ValueError("Invalid stat_type error. stat_type should be one of 'min', 'max', 'median', 'duration'")

    if stat_type is not None:
        template_dict["rasterFunctionArguments"]['ArgStatisticsType'] = stat_types[stat_type.lower()]

    if stat_type.lower() == 'duration':
        if min_value is None or max_value is None:
            raise ValueError("Missing min_value or max_value")
        else:
            template_dict["rasterFunctionArguments"]['MinValue'] = min_value
            template_dict["rasterFunctionArguments"]['MaxValue'] = max_value
    else:
        template_dict["rasterFunctionArguments"]['UndefinedClass'] = multiple_occurrence_value

    return Apply(in_raster=rasters,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def Aspect(raster):
    """
    Aspect identifies the downslope direction of the maximum rate of change in value from each cell to its neighbors.
    Aspect can be thought of as the slope direction. The values of the output raster will be the compass direction of
    the Aspect.

    Parameters:
      raster(Raster): Input raster

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "Aspect"
    }

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"])


def BandArithmetic(raster, band_ids, method=1):
    """

    The BandArithmetic function performs an arithmetic operation on the bands of a raster.

    Parameters:
      raster(Raster): Input raster
      band_ids(String): band ids (one-based) seperated by space
      method(int): Default 1. The type of band arithmetic algorithm you want to deploy.
                   You can define your custom algorithm, or choose a predefined index.
                   0 = UserDefined,
                   1 = NDVI,
                   2 = SAVI,
                   3 = TSAVI,
                   4 = MSAVI,
                   5 = GEMI,
                   6 = PVI,
                   7 = GVITM,
                   8 = Sultan,
                   9 = VARI,
                   10 = GNDVI,
                   11 = SR,
                   12 = NDVIre,
                   13 = SRre,
                   14 = MTVI2,
                   15 = RTVICore,
                   16 = CIre,
                   17 = CIg,
                   18 = NDWI,
                   19 = EVI,
                   20 = IronOxide,
                   21 = FerrousMinerals,
                   22 = ClayMinerals
                   23 = WNDWI

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "BandArithmetic",
        "rasterFunctionArguments": {
            "Method": method,
            "BandIndexes": band_ids
        }
    }

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def CIg(raster, nir_band_id=7, green_band_id=3):
    """
    The Chlorophyll Index - Green (CIg) is a vegetation index for estimating
    the chlorophyll content in leaves using the ratio of reflectivity in
    the near-infrared (NIR) and green bands.

    CIg = [(NIR / Green)-1]

    Parameters:
      raster(Raster): Input raster
      nir_band_id(int): NIR band ID, e.g., 7. Note that band ID uses one-based indexing.
      green_band_id(int): green band ID, e.g., 3. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # type check
    band_id_list = [nir_band_id, green_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int):
            raise ValueError("band id should be positive integer value")
        if band_id <= 0:
            raise ValueError("Invalid band id value(s). Note that band id value starts from 1 instead of 0")

    band_ids = str(str(nir_band_id) + " " + str(green_band_id))
    return BandArithmetic(raster, band_ids, 17)


def CIre(raster, nir_band_id=7, redEdge_band_id=6):
    """
    The Chlorophyll Index - Red-Edge (CIre) is a vegetation index for estimating
    the chlorophyll content in leaves using the ratio of reflectivity in the
    near-infrared (NIR) and red-edge bands.

    CIre = [(NIR / RedEdge)-1]

    Parameters:
      raster(Raster): Input raster
      nir_band_id(int): NIR band ID, e.g., 7. Note that band ID uses one-based indexing.
      redEdge_band_id(int): red-edge band ID, e.g., 6. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # type check
    band_id_list = [nir_band_id, redEdge_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(str(nir_band_id) + " " + str(redEdge_band_id))
    return BandArithmetic(raster, band_ids, 16)


def ClayMinerals(raster, swir1_band_id=6, swir2_band_id=7):
    """
    The Clay Minerals (CM) ratio is a geological index for identifying
    mineral features containing clay and alunite using two shortwave
    infrared (SWIR) bands. CM is used in mineral composite mapping.

    CM = SWIR1 / SWIR2

    Parameters:
      raster(Raster): Input raster
      swir1_band_id(int): SWIR1 band ID, e.g., 6. Note that band ID uses one-based indexing.
      swir2_band_id(int): SWIR2 band ID, e.g., 7. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    band_id_list = [swir1_band_id, swir2_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(str(swir1_band_id) + " " + str(swir2_band_id))
    return BandArithmetic(raster, band_ids, 22)


def ColorspaceConversion(raster, conversion_type="rgb_to_hsv"):
    """
    The ColorspaceConversion function converts the color model of a three-band
    unsigned 8-bit image from either the hue, saturation, and value (HSV)
    to red, green, and blue (RGB) or vice versa. An ExtractBand function and/or
    a Stretch function are sometimes used for converting the imagery into valid
    input of ColorspaceConversion function.

    Parameters:
      raster(Raster): Input raster. The pixeltype should be unsigned 8-bit
      conversion_type(string):  one of "rgb_to_hsv" or "hsv_to_rgb". Default is "rgb_to_hsv"

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    conversion_types = {
        "rgb_to_hsv": 0,
        "hsv_to_rgb": 1
    }

    if conversion_type not in conversion_types.keys():
        raise ValueError("conversion_type value is invalid. It should be either rgb_to_hsv or hsv_to_rgb")

    template_dict = {
        "rasterFunction": "ColorspaceConversion",
        "rasterFunctionArguments": {}
    }

    template_dict["rasterFunctionArguments"]['ConversionType'] = conversion_types[conversion_type]

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def ColormapToRGB(raster):
    """
    The function is designed to work with single band image service that has
    internal colormap. It will convert the image into a three-band 8-bit RGB
    raster. This function takes no arguments except an input raster.

    Parameters:
      raster(Raster): Input raster

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "ColormapToRGB"
    }

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"])


def Colormap(raster, colormap=None, colorramp=None):
    """
    Transforms the pixel values to display the raster data as a color (RGB) image, based on specific colors in
    a color map.

    Parameters:
      raster(Raster): Input raster
      colormap(string or list or dict): Default value is None.
                        Could be a string specifying the color map name, One of Random | NDVI | Elevation | Gray
                        Could be a list in the form of
                            [
                               [<value1>, <red1>, <green1>, <blue1>], //[int, int, int, int]
                               [<value2>, <red2>, <green2>, <blue2>],
                               ...
                            ]
                        Could be a dict in the form of:
                            {"values":[<value_1>, <value_2>,...], "colors":[<color_1>, <color_2>,...], "labels":[<label_1>, <label_2>,...]}
                            here, labels are optional

      colorramp(string or dict): Default value is None.
                        Can be a string specifying color ramp name like "Black To White", "Yellow To Red", "Slope", or other colorramp names supported in ArcGIS Pro
                        Can be a dict. For more information about colorramp object, see color ramp object https://developers.arcgis.com/documentation/common-data-types/color-ramp-objects.htm
    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """



    template_dict = {
        "rasterFunction": "Colormap",
        "rasterFunctionArguments": {}
    }

    if colormap is not None and isinstance(colormap, str):
        template_dict["rasterFunctionArguments"]['ColormapName'] = colormap
    if colormap is not None and (isinstance(colormap, list) or isinstance(colormap, dict)):
        template_dict["rasterFunctionArguments"]['Colormap'] = colormap
    if colorramp is not None and isinstance(colorramp, str):
        template_dict["rasterFunctionArguments"]['ColorrampName'] = colorramp
    if colorramp is not None and isinstance(colorramp, dict):
        template_dict["rasterFunctionArguments"]['Colorramp'] = colorramp
    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def Complex(raster):
    """
    Complex function computes magnitude from complex values. It is used when
    input raster has complex pixel type. It computes magnitude from complex
    value to convert the pixel type to floating point for each pixel. It takes
    no argument but an input raster.

    Parameters:
      raster(Raster): Input raster

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "Complex"
    }

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"])


def CompositeBand(rasters):
    """
    Combines multiple images to form a multiband image.

    Parameters:
      raster(list of Rasters): Input rasters

    Returns:
      output_raster(Raster): The output multiband raster.
    """

    template_dict = {
        "rasterFunction": "CompositeBand"
    }

    return Apply(in_raster=rasters,
                 raster_function=template_dict["rasterFunction"])


def ContrastBrightness(raster, contrast_offset=0, brightness_offset=0):
    """
    The ContrastBrightness function enhances the appearance of raster data (imagery) by modifying the brightness or
    contrast within the image. This function works on 8-bit input raster only.

    Parameters:
      raster(Raster): Input raster. The pixel type of the input raster must be 8-bit .
      contrast_offset(float): contrast offset value. Value can be in range -100 to 100
      brightness_offset(float): brightness offset value. Value can be in range -100 to 100

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    from numbers import Number
    # type check
    if not isinstance(contrast_offset, Number) or contrast_offset < -100 or contrast_offset > 100:
        raise ValueError("contrast_offset should be a number within [-100, 100]")
    if not isinstance(brightness_offset, Number) or brightness_offset < -100 or brightness_offset > 100:
        raise ValueError("brightness_offset should be a number within [-100, 100]")

    template_dict = {
        "rasterFunction": "ContrastBrightness",
        "rasterFunctionArguments": {
            "ContrastOffset": contrast_offset,
            "BrightnessOffset": brightness_offset
        }
    }

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def Convolution(raster, kernel=None):
    """
    The Convolution function performs filtering on the pixel values in an image, which can be used for sharpening an
    image, blurring an image, detecting edges within an image, or other kernel-based enhancements.

    Parameters:
      raster(Raster): Input raster
      kernel(int, list of list): well known kernel or user defined kernel passed as a list of list.
                                 well known kernels:
                                 LINE_DETECTION_HORIZONTAL = 0
                                 LINE_DETECTION_VERTICAL = 1
                                 LINE_DETECTION_LEFT_DIAGONAL = 2
                                 LINE_DETECTION_RIGHT_DIAGONAL = 3
                                 GRADIENT_NORTH = 4
                                 GRADIENT_WEST = 5
                                 GRADIENT_EAST = 6
                                 GRADIENT_SOUTH = 7
                                 GRADIENT_NORTH_EAST = 8
                                 GRADIENT_NORTH_WEST = 9
                                 SMOOTH_ARITHMETIC_MEAN = 10
                                 SMOOTHING_3X3 = 11
                                 SMOOTHING_5X5 = 12
                                 SHARPENING_3X3 = 13
                                 SHARPENING_5X5 = 14
                                 LAPLACIAN_3X3 = 15
                                 LAPLACIAN_5X5 = 16
                                 SOBEL_HORIZONTAL = 17
                                 SOBEL_VERTICAL = 18
                                 SHARPEN = 19
                                 SHARPEN2 = 20
                                 POINT_SPREAD = 21

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "Convolution",
        "rasterFunctionArguments": {}
    }

    if (isinstance(kernel, int)):
        template_dict["rasterFunctionArguments"]['Type'] = kernel
    elif (isinstance(kernel, list)):
        numrows = len(kernel)
        numcols = len(kernel[0])
        flattened = [item for sublist in kernel for item in sublist]
        template_dict["rasterFunctionArguments"]['Columns'] = numcols
        template_dict["rasterFunctionArguments"]['Rows'] = numrows
        template_dict["rasterFunctionArguments"]['Kernel'] = flattened
        template_dict["rasterFunctionArguments"]['Type'] = -1
    else:
        raise ValueError('Invalid kernel type - pass int or list of list: [[][][]...]')

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def Curvature(raster, curvature_type='standard', z_factor=1):
    """
    The Curvature function displays the shape or curvature of the slope. A part of a surface can be concave or convex;
    you can tell that by looking at the curvature value. The curvature is calculated by computing the second derivative
    of the surface.

    Parameters:
      raster(Raster): Input raster
      curvature_type(string): default value is 'standard'. The curvature type accentuates different aspects of the slope.
                              One of 'standard', 'planform', 'profile'
                              standard - Combines both the Profile and Planform curvatures.

                              planform - Is parallel to the slope and indicates the direction of maximum slope.
                              It affects the acceleration and deceleration of flow across the surface.

                              profile - Is perpendicular to the direction of the maximum slope.
                              It affects the convergence and divergence of flow across a surface.

      z_factor(double): The z-factor adjusts the units of measure for the z units when they are
                        different from the x,y units of the input surface. If the x,y units and
                        z units are in the same units of measure, the z-factor should be set to 1.
                        The z-values of the input surface are multiplied by the z-factor when
                        calculating the final output surface. For example, if your z units are
                        feet and your x,y units are meters, you would use a z-factor of 0.3048

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    curv_types = {
        'standard': 0,
        'profile': 1,
        'planform': 2
    }
    # value check
    if curvature_type.lower() not in curv_types.keys():
        raise ValueError(
            "Invalid curvature_type value. Note that curvature_type should be one of 'standard', 'planform', 'profile'")

    in_curv_type = curv_types[curvature_type.lower()]

    template_dict = {
        "rasterFunction": "Curvature",
        "rasterFunctionArguments": {
            "Type": in_curv_type,
            "ZFactor": z_factor
        }
    }

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def ElevationVoidFill(raster, max_void_width=0):
    """
    The elevation_void_fill function is used to create pixels where holes exist in your elevation.

    Parameters:
      raster(Raster): Input raster
      max_void_width(int): Default is 0. The maximum void width value is used to specify the largest size of a void
                            that you want to fill. If the width or height of the bounding box around the
                            void is larger than the maximum void width value, the void is not filled.
                            The units of this parameter is the same as the units used in your data's
                            spatial reference system

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "ElevationVoidFill",
        "rasterFunctionArguments": {}
    }

    if max_void_width is not None:
        template_dict["rasterFunctionArguments"]["MaxVoidWidth"] = max_void_width

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def ExtractBand(raster, band_ids=None, band_names=None, band_wavelengths=None, missing_band_action="BestMatch",
                wavelength_match_tolerance=None):
    """
    The ExtractBand function allows you to extract one or more bands from a raster, or it can reorder the bands in a
    multiband image.

    Parameters:
      raster(Raster): Input raster
      band_ids(int or list of int): single band ID or list of band IDs. Note band ID (one-based indexing)
      band_names(string or list of string): single band name or list of Band names
      band_wavelengths(list of double): Band wavelengths
      missing_band_action(int): 0 = BestMatch, 1 = Fail
      wavelength_match_tolerance(double): wavelength match tolerance

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # value check
    missing_band_actions = {"BestMatch": 0, "Fail": 1}

    if missing_band_action not in missing_band_actions.keys():
        raise ValueError(
            "missing_band_action value is invalid. Note that missing_band_action should be either 'BestMatch' or 'Fail'")

    template_dict = {
        "rasterFunction": "ExtractBand",
        "rasterFunctionArguments": {}
    }

    if band_ids is not None:
        if isinstance(band_ids, list):
            for item in band_ids:
                if not isinstance(item, int) or item <= 0:
                    raise ValueError("band id should be positive integer value")
            template_dict["rasterFunctionArguments"]["BandIDs"] = band_ids
        elif isinstance(band_ids, int):
            if band_ids <= 0:
                raise ValueError("band id should be positive integer value")
            template_dict["rasterFunctionArguments"]["BandIDs"] = [band_ids]
        else:
            raise TypeError("band_ids should be a single integer or a list of interger values")
        template_dict["rasterFunctionArguments"]['Method'] = 2

    if band_names is not None:
        if isinstance(band_names, list):
            for item in band_names:
                if not isinstance(item, str):
                    raise ValueError("band name should be string")
            template_dict["rasterFunctionArguments"]["BandNames"] = band_names
        elif isinstance(band_names, str):
            template_dict["rasterFunctionArguments"]["BandNames"] = [band_names]
        else:
            raise TypeError("band_names should be  either a single string or a list of string")

    if band_wavelengths is not None:
        template_dict["rasterFunctionArguments"]["BandWavelengths"] = band_wavelengths

    if missing_band_action is not None:
        template_dict["rasterFunctionArguments"]["MissingBandAction"] = missing_band_actions[missing_band_action]
    if wavelength_match_tolerance is not None:
        template_dict["rasterFunctionArguments"]["WavelengthMatchTolerance"] = wavelength_match_tolerance

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def EVI(raster, nir_band_id=5, red_band_id=4, blue_band_id=2):
    """
    The Enhanced Vegetation Index (EVI) is an optimized vegetation index that accounts
    for atmospheric influences and vegetation background signal. It's similar to NDVI,
    but is less sensitive to background and atmospheric noise, and it does not become
    saturated NDVI when viewing areas with very dense green vegetation.

    EVI =  2.5 * [(NIR - Red)/(NIR + (6*Red) - (7.5*Blue) + 1)]

    Parameters:
      raster(Raster): Input raster
      nir_band_id(int): NIR band ID, e.g., 5. Note that band ID uses one-based indexing.
      red_band_id(int): red band ID, e.g., 4. Note that band ID uses one-based indexing.
      blue_band_id(int): blue band ID , e.g., 2. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # type check
    band_id_list = [nir_band_id, red_band_id, blue_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")

    band_ids = str(nir_band_id) + " " + str(red_band_id) + " " + str(blue_band_id)
    return BandArithmetic(raster, band_ids, 19)


def FerrousMinerals(raster, swir_band_id=6, nir_band_id=5):
    """
    The Ferrous Minerals (FM) ratio is a geological index for identifying
    rock features containing some quantity of iron-bearing minerals using
    the shortwave infrared (SWIR) and near-infrared (NIR) bands. FM is used
    in mineral composite mapping.

    FM = SWIR / NIR

    Parameters:
      raster(Raster): Input raster
      swir_band_id(int): SWIR band ID, e.g., 6. Note that band ID uses one-based indexing.
      nir_band_id(int): NIR band ID, e.g., 5. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # type check
    band_id_list = [swir_band_id, nir_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(str(swir_band_id) + " " + str(nir_band_id))
    return BandArithmetic(raster, band_ids, 21)


def Statistics(raster, kernel_columns=3, kernel_rows=3, stat_type=None, fill_no_data_only=False):
    """
    Function calculates focal statistics for each pixel of an image based on a defined focal neighborhood.

    Parameters:
      raster(Raster): Input raster
      kernel_columns(int): The number of pixel columns to use in your focal neighborhood dimension.
      kernel_rows(int): The number of pixel rows to use in your focal neighborhood dimension
      stat_type(string): There are four types of focal statistical functions:
                       Min, Max, Mean, StandardDeviation
                          -Min-Calculates the minimum value of the pixels within the neighborhood
                          -Max-Calculates the maximum value of the pixels within the neighborhood
                          -Mean-Calculates the average value of the pixels within the neighborhood. This is the default.
                          -StandardDeviation-Calculates the standard deviation value of the pixels within the neighborhood
      fill_no_data_only(bool): Set to True, to fill no data pixels only.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    statistics_types = {"min": 1, "max": 2, "mean": 3, "standarddeviation": 4}

    template_dict = {
        "rasterFunction": "Statistics",
        "rasterFunctionArguments": {}
    }

    if stat_type is not None:
        if isinstance(stat_type, str) and stat_type.lower() in statistics_types:
            template_dict["rasterFunctionArguments"]['Type'] = statistics_types[stat_type]
        else:
            raise ValueError(
                "Invalid stat_type value. Note that stat_type must be one of min, max, mean, standarddeviation")

    if kernel_columns is not None:
        template_dict["rasterFunctionArguments"]["Columns"] = kernel_columns
    if kernel_rows is not None:
        template_dict["rasterFunctionArguments"]["Rows"] = kernel_rows
    if fill_no_data_only is not None:
        template_dict["rasterFunctionArguments"]["FillNoDataOnly"] = fill_no_data_only

    return Apply(in_raster=raster,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def Hillshade(dem, azimuth=315.0, altitude=45.0, z_factor=1, slope_type=1, ps_power=0.664, psz_factor=0.024,
              remove_edge_effect=False, hillshade_type=0):
    """
    A hillshade is a grayscale 3D model of the surface taking the sun's relative position into account to shade the image.

    Parameters:
      dem(Raster): Input DEM raster
      azimuth(double): default 315.0 Azimuth is the sun's relative position along the horizon (in degrees).
                       This position is indicated by the angle of the sun measured clockwise from due north.
                       An azimuth of 0 degrees indicates north, east is 90 degrees, south is 180 degrees,
                       and west is 270 degrees.
      altitude(double): default 45.0 Altitude is the sun's angle of elevation above the horizon and ranges from 0 to 90 degrees.
                        A value of 0 degrees indicates that the sun is on the horizon, that is, on the same horizontal
                        plane as the frame of reference. A value of 90 degrees indicates that the sun is directly overhead.
      z_factor(double): default 1. The z-factor is a scaling factor used to convert the elevation values for two purposes:

                        - Convert the elevation units (such as meters or feet) to the horizontal coordinate units of the dataset,
                          which may be feet, meters, or degrees.

                        - Add vertical exaggeration for visual effect.

      slope_type(int): default is 1. 1=DEGREE, 2=PERCENTRISE, 3=SCALED.
      ps_power(double):  default value is 0.664. used together with SCALED slope type. Pixel Size Power accounts for the altitude changes (or scale)
                        as the viewer zooms in and out on the map display. It is the exponent applied to the pixel size term
                        in the equation that controls the rate at which the Z Factor changes to avoid significant loss of
                        relief.
      psz_factor(double):  default value is 0.024. used together with SCALED slope type. Pixel Size Factor accounts for changes in scale as the viewer
                         zooms in and out on the map display. It controls the rate at which the Z Factor changes.
      remove_edge_effect(bool): Set to True to remove edge effect
      hillshade_type(int): Default is 0, 0 = traditional, 1 = multi - directional

                           Traditional - Calculates hillshade from a single illumination direction.
                           You can set the Azimuth and Altitude to control the location of the light source.

                           Multidirectional - Combines light from multiple sources to represent an enhanced
                           visualization of the terrain.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "Hillshade",
        "rasterFunctionArguments": {}
    }

    if azimuth is not None:
        template_dict["rasterFunctionArguments"]["Azimuth"] = azimuth
    if altitude is not None:
        template_dict["rasterFunctionArguments"]["Altitude"] = altitude
    if z_factor is not None:
        template_dict["rasterFunctionArguments"]["ZFactor"] = z_factor
    if slope_type is not None:
        template_dict["rasterFunctionArguments"]["SlopeType"] = slope_type
    if ps_power is not None:
        template_dict["rasterFunctionArguments"]["PSPower"] = ps_power
    if psz_factor is not None:
        template_dict["rasterFunctionArguments"]["PSZFactor"] = psz_factor
    if remove_edge_effect is not None:
        template_dict["rasterFunctionArguments"]["RemoveEdgeEffect"] = remove_edge_effect
    if hillshade_type is not None:
        template_dict["rasterFunctionArguments"]["HillshadeType"] = hillshade_type

    return Apply(in_raster=dem,
                 raster_function=template_dict["rasterFunction"],
                 raster_function_arguments=template_dict["rasterFunctionArguments"])


def GEMI(raster, nir_band_id=4, red_band_id=3):
    """
    Global Environmental Monitoring Index
    GEMI = eta*(1-0.25*eta)-((Red-0.125)/(1-Red))
    where eta = (2*(NIR^2-Red^2)+1.5*NIR+0.5*Red)/(NIR+Red+0.5)

    Parameters:
      raster(Raster): Input raster
      nir_band_id(int): NIR band ID, e.g., 4. Note that band ID uses one-based indexing.
      red_band_id(int): red band ID, e.g., 3. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # type check
    band_id_list = [nir_band_id, red_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(str(nir_band_id) + " " + str(red_band_id))
    return BandArithmetic(raster, band_ids, 5)


def GNDVI(raster, nir_band_id=4, green_band_id=2):
    """
    Green Normalized Difference Vegetation Index

    GNDVI = (NIR-Green)/(NIR+Green)

    Parameters:
      raster(Raster): Input raster
      nir_band_id(int): NIR band ID, e.g., 4. Note that band ID uses one-based indexing.
      green_band_id(int): green band ID, e.g., 2. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # type check
    band_id_list = [nir_band_id, green_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(str(nir_band_id) + " " + str(green_band_id))
    return BandArithmetic(raster, band_ids, 10)


def GVITM(raster, band1_id=1, band2_id=2, band3_id=3, band4_id=4, band5_id=5, band7_id=7):
    """
    Green Vegetation Index - Landsat TM
    GVITM = -0.2848*Band1-0.2435*Band2-0.5436*Band3+0.7243*Band4+0.0840*Band5-0.1800*Band7

    Parameters:
      :param raster(Raster): Input raster
      :param band1_id(int): the band ID of band1, e.g., 1. Note that band ID uses one-based indexing.
      :param band2_id(int): the band ID of band2, e.g., 2. Note that band ID uses one-based indexing.
      :param band3_id(int): the band ID of band3, e.g., 3. Note that band ID uses one-based indexing.
      :param band4_id(int): the band ID of band4, e.g., 4. Note that band ID uses one-based indexing.
      :param band5_id(int): the band ID of band5, e.g., 5. Note that band ID uses one-based indexing.
      :param band7_id(int): the band ID of band6, e.g., 7. Note that band ID uses one-based indexing.

    Returns:
      output_raster(Raster): The output raster with this function applied to it
    """
    # type check
    band_id_list = [band1_id, band2_id, band3_id, band4_id, band5_id, band7_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(
        str(band1_id) + " " + str(band2_id) + " " + str(band3_id) + " " + str(band4_id) + " " + str(
            band5_id) + " " + str(band7_id))
    return BandArithmetic(raster, band_ids, 7)


def IronOxide(raster, red_band_id=4, blue_band_id=2):
    """
    The Iron Oxide (IO) ratio is a geological index for identifying rock
    features that have experienced oxidation of iron-bearing sulfides
    using the red and blue bands. IO is useful in identifying iron oxide
    features below vegetation canopies, and is used in mineral composite mapping.

    IronOxide = Red / Blue

    :param raster: the input raster / imagery layer
    :param red_band_id: the band ID of the red band, e.g.,4. Note that band ID uses one-based indexing.
    :param blue_band_id: the band ID of the blue band, e.g.,2. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [red_band_id, blue_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(red_band_id) + " " + str(blue_band_id)
    return BandArithmetic(raster, band_ids, 20)


def Lookup(raster, field=None):
    """
    Creates a new raster by looking up values found in another field in the table of the input raster.
    For more information see, https://pro.arcgis.com/en/pro-app/help/data/imagery/lookup-function.htm

    :param raster: The input raster that contains a field from which to create a new raster.
    :param field: Field containing the desired values for the new raster.

    :return: the output raster with this function applied to it
    """

    template_dict = {
        "rasterFunction": "Lookup",
        "rasterFunctionArguments": {
        }
    }

    if field is not None:
        template_dict["rasterFunctionArguments"]['Field'] = field

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Mask(raster, no_data_values=None, included_ranges=None, no_data_interpretation=None):
    """
    The mask function changes the image by specifying a certain pixel value or a range of pixel values as no data.
    The arguments for the mask function are as follows:

    :param raster: input raster
    :param no_data_values: array of numbers [band0_val,band1_val,...]
    :param included_ranges: array of double [band0_lowerbound,band0_upperbound,band1...],
    :param no_data_interpretation: int 0=MatchAny, 1=MatchAll
    :return: the output raster with this function applied to it

    """

    template_dict = {
        "rasterFunction": "Mask",
        "rasterFunctionArguments": {
        }
    }

    if no_data_values is not None:
        template_dict["rasterFunctionArguments"]["NoDataValues"] = no_data_values
    if included_ranges is not None:
        template_dict["rasterFunctionArguments"]["IncludedRanges"] = included_ranges
    if no_data_interpretation is not None:
        template_dict["rasterFunctionArguments"]["NoDataInterpretation"] = no_data_interpretation

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def MLClassify(raster, signature):
    """
    The ml_classify function allows you to perform a supervised classification using the maximum likelihood classification
     algorithm. For more information, see https://pro.arcgis.com/en/pro-app/help/data/imagery/ml-classify-function.htm
     The arguments for the ml_classify function are as follows:

    :param raster: input raster
    :param signature: string. a signature string returned from computeClassStatistics (GSG)
    :return: the output raster

    """

    template_dict = {
        "rasterFunction": "MLClassify",
        "rasterFunctionArguments": {
            "SignatureFile": signature
        }
    }

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def MSAVI(raster, nir_band_id=4, red_band_id=3):
    """
    Modified Soil Adjusted Vegetation Index
    MSAVI2 = (1/2)*(2(NIR+1)-sqrt((2*NIR+1)^2-8(NIR-Red)))

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,4. Note that band ID uses one-based indexing.
    :param red_band_id: the band ID of the red band, e.g.,3. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [nir_band_id, red_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(red_band_id)
    return BandArithmetic(raster, band_ids, 4)


def MTVI2(raster, nir_band_id=7, red_band_id=5, green_band_id=3):
    """
    The Modified Triangular Vegetation Index (MTVI2) is a vegetation index
    for detecting leaf chlorophyll content at the canopy scale while being
    relatively insensitive to leaf area index. It uses reflectance in the green,
    red, and near-infrared (NIR) bands

    MTVI2 = (1.5*(1.2*(NIR-Green)-2.5*(Red-Green))/sqrt((2*NIR+1)^2-(6*NIR-5*sqrt(Red))-0.5))

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,7. Note that band ID uses one-based indexing.
    :param red_band_id: the band ID of the red band, e.g.,5. Note that band ID uses one-based indexing.
    :param green_band_id: the band ID of the green band, e.g.,3. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [nir_band_id, red_band_id, green_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(red_band_id) + " " + str(green_band_id)
    return BandArithmetic(raster, band_ids, 14)


def NDVI(raster, nir_band_id=4, red_band_id=3):
    """
    Normalized Difference Vegetation Index
    NDVI = ((NIR - Red)/(NIR + Red))

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,4. Note that band ID uses one-based indexing. .
    :param red_band_id: the band ID of the red band, e.g.,3. Note that band ID uses one-based indexing. .
    :return: Normalized Difference Vegetation Index raster
    """
    band_id_list = [nir_band_id, red_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(red_band_id)
    return BandArithmetic(raster, band_ids, 1)


def NDVIre(raster, nir_band_id=7, redEdge_band_id=6):
    """
    Red-Edge NDVI (NDVIre)
    The Red-Edge NDVI (NDVIre) is a vegetation index for estimating
    vegetation health using the red-edge band. It is especially useful
    for estimating crop health in the mid to late stages of growth where
    the chlorophyll concentration is relatively higher. Also, NDVIre can
    be used to map the within-field variability of nitrogen foliage to
    understand the fertilizer requirements of crops.

    NDVIre = (NIR-RedEdge)/(NIR+RedEdge)

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,7. Note that band ID uses one-based indexing.
    :param rededge_band_id: the band ID of the redEdge band, e.g.,6. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [nir_band_id, redEdge_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(redEdge_band_id)
    return BandArithmetic(raster, band_ids, 12)


def NDWI(raster, nir_band_id=5, green_band_id=3):
    """
    The Normalized Difference Water Index (NDWI) is an index for delineating and
    monitoring content changes in surface water. It is computed with the near-infrared
    (NIR) and green bands.

    NDWI = (Green - NIR)/(Green +NIR)

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,5. Note that band ID uses one-based indexing.
    :param green_band_id: the band ID of the green band, e.g.,3. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [nir_band_id, green_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(green_band_id)
    return BandArithmetic(raster, band_ids, 18)


def Pansharpen(pan_raster, ms_raster, ir_raster=None, fourth_band_of_ms_is_ir=True, weights=[0.166, 0.167, 0.167, 0.5],
               type="ESRI", sensor=None):
    """
    The Pansharpening function uses a higher-resolution panchromatic raster to
    fuse with a lower-resolution, multiband raster. It can generate colorized
    multispectral image with higher resolution. For more information, see
    https://pro.arcgis.com/en/pro-app/help/data/imagery/pansharpening-function.htm

    :param pan_raster: raster, which is panchromatic
    :param ms_raster: raster, which is multispectral
    :param ir_raster: Optional, if fourth_band_of_ms_is_ir is true or selected pansharpening method doesn't require near-infrared image
    :param fourth_band_of_ms_is_ir: Boolean, "true" if "ms_raster" has near-infrared image on fourth band
    :param weights: Weights applied for Red, Green, Blue, Near-Infrared bands. 4-elements array, Sum of values is 1
    :param type: string, describes the Pansharpening method one of "IHS", "Brovey" "ESRI", "SimpleMean", "Gram-Schmidt". Default is "ESRI"
    :param sensor: string, it is an optional parameter to specify the sensor name
    :return: output raster with function applied
    """

    pansharpening_types = {
        "IHS": 0,
        "Brovey": 1,
        "ESRI": 2,
        "SimpleMean": 3,
        "Gram-Schmidt": 4
    }

    template_dict = {
        "rasterFunction": "Pansharpening",
        "rasterFunctionArguments": {
            "Weights": weights
        }
    }

    if type is not None:
        template_dict["rasterFunctionArguments"]['PansharpeningType'] = pansharpening_types[type]

    if isinstance(fourth_band_of_ms_is_ir, bool):
        template_dict["rasterFunctionArguments"]['UseFourthBandOfMSAsIR'] = fourth_band_of_ms_is_ir

    if sensor is not None:
        template_dict["rasterFunctionArguments"]['Sensor'] = sensor

    if ir_raster is not None:
        return Apply({"PanImage": pan_raster, "MSImage": ms_raster, "InfraredImage": ir_raster},
                     template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])
    else:
        return Apply({"PanImage": pan_raster, "MSImage": ms_raster}, template_dict['rasterFunction'],
                     template_dict['rasterFunctionArguments'])


def PVI(raster, nir_band_id=4, red_band_id=3, a=0.3, b=0.5):
    """
    Perpendicular Vegetation Index
    PVI = (NIR-a*Red-b)/(sqrt(1+a^2))

    :param raster: the input raster
    :param nir_band_id: the band ID of the nir band, e.g.,4. Note that band ID uses one-based indexing.
    :param red_band_id: the band ID of the red band, e.g.,3. Note that band ID uses one-based indexing.
    :param a: the slope
    :param b: the gradient

    :return: output raster
    """
    band_id_list = [nir_band_id, red_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(red_band_id) + " " + str(a) + " " + str(b)
    return BandArithmetic(raster, band_ids, 6)


def RasterCalculator(rasters, input_names, expression, extent_type="FirstOf", cellsize_type="FirstOf"):
    """
    The RasterCalculator function provides access to all existing math functions
    so you can make calls to them when building your expressions. The calculator
    function requires single-band inputs. If you need to perform expressions on
    bands in a multispectral image as part of a function chain, you can use
    the Extract Bands Function before the RasterCalculator function.
    For more info including operators supported, see Calculator function
    http://pro.arcgis.com/en/pro-app/help/data/imagery/calculator-function.htm

    :param raster: array of rasters
    :param input_names: array of strings for arbitrary raster names.
    :param expression: string, expression to calculate output raster from input rasters
    :param extent_type: string, one of "FirstOf", "IntersectionOf" "UnionOf", "LastOf". Default is "FirstOf".
    :param cellsize_type: one of "FirstOf", "MinOf", "MaxOf "MeanOf", "LastOf". Default is "FirstOf".
    :return: output raster with function applied
    """

    extent_types = {
        "FirstOf": 0,
        "IntersectionOf": 1,
        "UnionOf": 2,
        "LastOf": 3
    }

    cellsize_types = {
        "FirstOf": 0,
        "MinOf": 1,
        "MaxOf": 2,
        "MeanOf": 3,
        "LastOf": 4
    }

    template_dict = {
        "rasterFunction": "RasterCalculator",
        "rasterFunctionArguments": {
            "InputNames": input_names,
            "Expression": expression
        },
        "variableName": "Rasters"
    }

    template_dict["rasterFunctionArguments"]['ExtentType'] = extent_types[extent_type]
    template_dict["rasterFunctionArguments"]['CellsizeType'] = cellsize_types[cellsize_type]

    return Apply(rasters, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Remap(raster, input_ranges=None, output_values=None, no_data_ranges=None, allow_unmatched=False):
    """
    The remap function allows you to change or reclassify the pixel values of the raster data.
    For more information, see https://pro.arcgis.com/en/pro-app/help/data/imagery/remap-function.htm

    The arguments for the remap function are as follows:

    :param raster: input raster
    :param input_ranges: [<range1_from>, <range1_to>, …], input ranges are specified in pairs: from (inclusive) and to (exclusive). Default: None
    :param output_values: [new_value1, ...], output values of corresponding input ranges. Default: None
    :param no_data_ranges: [<nodatavalue1_from>, < nodatavalue1_to>, …],//[double, double, …], nodata ranges are specified in pairs: from (inclusive) and to (exclusive). Default: None
    :param allow_unmatched: Boolean, specify whether to keep the unmatched values or turn into nodata. Default: False
    :return: the remapped raster

    """

    template_dict = {
        "rasterFunction": "Remap",
        "rasterFunctionArguments": {
        },
        "variableName": "Raster"
    }

    if input_ranges is None and output_values is None and no_data_ranges is None:
        raise ValueError("input_ranges, output_values and no_data_ranges should not be None at the same time")

    if input_ranges is not None:
        template_dict["rasterFunctionArguments"]["InputRanges"] = input_ranges
        if output_values is not None:
            template_dict["rasterFunctionArguments"]["OutputValues"] = output_values
        else:
            raise ValueError("output_values should be provided.")
    if no_data_ranges is not None:
        template_dict["rasterFunctionArguments"]["NoDataRanges"] = no_data_ranges
    if allow_unmatched is not None:
        template_dict["rasterFunctionArguments"]["AllowUnmatched"] = allow_unmatched

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Resample(raster, resampling_type=None, input_cellsize=None, output_cellsize=None):
    """
    The resample function resamples pixel values from a given resolution.The arguments for the resample function are as follows:

    :param raster: input raster
    :param resampling_type: one of NearestNeighbor,Bilinear,Cubic,Majority,BilinearInterpolationPlus,BilinearGaussBlur,
            BilinearGaussBlurPlus, Average, Minimum, Maximum,VectorAverage(require two bands)
    :param input_cellsize: point that defines cellsize in source spatial reference
    :param output_cellsize: point that defines output cellsize
    :return: the output raster

    """

    resample_types = {
        'NearestNeighbor': 0,
        'Bilinear': 1,
        'Cubic': 2,
        'Majority': 3,
        'BilinearInterpolationPlus': 4,
        'BilinearGaussBlur': 5,
        'BilinearGaussBlurPlus': 6,
        'Average': 7,
        'Minimum': 8,
        'Maximum': 9,
        'VectorAverage': 10
    }

    if isinstance(resampling_type, str):
        resampling_type = resample_types[resampling_type]

    template_dict = {
        "rasterFunction": "Resample",
        "rasterFunctionArguments": {
        },
        "variableName": "Raster"
    }
    if resampling_type is not None:
        template_dict["rasterFunctionArguments"]["ResamplingType"] = resampling_type
    if input_cellsize is not None:
        template_dict["rasterFunctionArguments"]["InputCellsize"] = input_cellsize
    if output_cellsize is not None:
        template_dict["rasterFunctionArguments"]["OutputCellsize"] = output_cellsize

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def RTVICore(raster, nir_band_id=7, redEdge_band_id=6, green_band_id=3):
    """
    The Red-Edge Triangulated Vegetation Index (RTVICore) is a vegetation index
    for estimating leaf area index and biomass. This index uses reflectance
    in the NIR, red-edge, and green spectral bands

    RTVICore = [100(NIR-RedEdge)-10(NIR-Green)]

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,7. Note that band ID uses one-based indexing.
    :param redEdge_band_id: the band ID of the redEdge band, e.g.,6. Note that band ID uses one-based indexing.
    :param green_band_id: the band ID of the green band, e.g.,3. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [nir_band_id, redEdge_band_id, green_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(redEdge_band_id) + " " + str(green_band_id)
    return BandArithmetic(raster, band_ids, 15)


def SAVI(raster, nir_band_id=4, red_band_id=3, l=0.33):
    """
    Soil-Adjusted Vegetation Index
    SAVI = ((NIR - Red) / (NIR + Red + L)) x (1 + L)
    where L represents amount of green vegetative cover, e.g., 0.5

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,4. Note that band ID uses one-based indexing.
    :param red_band_id: the band ID of the red band, e.g.,3. Note that band ID uses one-based indexing.
    :param L: the amount of green vegetative cover, e.g., 0.33
    :return: output raster
    """
    band_id_list = [nir_band_id, red_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + " " + str(red_band_id) + " " + str(l)
    return BandArithmetic(raster, band_ids, 2)


def ShadedRelief(raster, azimuth=315, altitude=45, z_factor=1, colormap=None, colorramp="Elevation #1", slope_type="DEGREE",
                 ps_power=0.664, psz_factor=0.024, remove_edge_effect=False):
    """
    Shaded relief is a color 3D model of the terrain, created by merging the images from the Elevation-coded and
    Hillshade methods. For more information, see https://pro.arcgis.com/en/pro-app/help/data/imagery/shaded-relief-function.htm

    The arguments for the shaded_relief function are as follows:
        :param raster: input raster
        :param azimuth (double): default 315
        :param altitude (double): default 45
        :param z_factor (double): default 1
        :param colormap(list or dict): default is None.
                                    Could be a list in the form of
                                        [
                                            [<value1>, <red1>, <green1>, <blue1>], //[int, int, int, int]
                                           [<value2>, <red2>, <green2>, <blue2>],
                                           ...
                                        ]
                                    Could be a dict in the form of:
                                        {"values":[<value_1>, <value_2>,...], "colors":[<color_1>, <color_2>,...], "labels":[<label_1>, <label_2>,...]}
                                    here, labels are optional
        :param colorramp(string or dict): default is "Elevation #1".
                                    Could be a string specifying colorramp name like "Black To White", "Yellow To Red", "Slope", or other colorramp names supported in ArcGIS Pro
                                    Could be a dict. For more information about colorramp object, see color ramp object https://developers.arcgis.com/documentation/common-data-types/color-ramp-objects.htm
        :param slope_type(string): one of the following "DEGREE", "PERCENTRISE", or "SCALED". Default is "DEGREE"
        :param ps_power(double): used when slope_type is SCALED. default 0.664
        :param psz_factor(double): used when slope_type is SCALED. default 0.024
        :param remove_edge_effect(boolean): True or False. default False
    :return: the output raster
    """
    slope_types = {
        "degree": 1,
        "percentrise": 2,
        "scaled": 3,
    }

    if slope_type.lower() not in slope_types:
        raise ValueError(
            "Invalid slope_type value. Note that slope_type should be one of 'DEGREE', 'PERCENTRISE', or 'SCALED'")

    template_dict = {
        "rasterFunction": "ShadedRelief",
        "rasterFunctionArguments": {},
        "variableName": "Raster"
    }

    if azimuth is not None:
        template_dict["rasterFunctionArguments"]["Azimuth"] = azimuth
    if altitude is not None:
        template_dict["rasterFunctionArguments"]["Altitude"] = altitude
    if z_factor is not None:
        template_dict["rasterFunctionArguments"]["ZFactor"] = z_factor
    if colormap is not None:
        template_dict["rasterFunctionArguments"]["ColorSchemeType"] = 0
        template_dict["rasterFunctionArguments"]["Colormap"] = colormap
    elif colorramp is not None:
        template_dict["rasterFunctionArguments"]["ColorSchemeType"] = 1
        template_dict["rasterFunctionArguments"]["ColorRamp"] = colorramp
    else:
        raise ValueError("Both colormap and colorramp values are None. Please provide one.")
    if slope_type is not None:
        template_dict["rasterFunctionArguments"]["SlopeType"] = slope_types[slope_type.lower()]
    if slope_type.lower() == "scaled" and ps_power is not None:
        template_dict["rasterFunctionArguments"]["PSPower"] = ps_power
    if slope_type.lower() == "scaled" and psz_factor is not None:
        template_dict["rasterFunctionArguments"]["PSZFactor"] = psz_factor
    if remove_edge_effect is not None:
        template_dict["rasterFunctionArguments"]["RemoveEdgeEffect"] = remove_edge_effect

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Slope(dem, z_factor=1, slope_type="DEGREE", ps_power=0.664, psz_factor=0.024, remove_edge_effect=False):
    """
    slope represents the rate of change of elevation for each pixel. For more information, see
    https://pro.arcgis.com/en/pro-app/help/data/imagery/slope-function.htm
    The arguments for the slope function are as follows:

    :param dem: input DEM
    :param z_factor: double, default is 1
    :param slope_type: one of the following values: DEGREE, PERCENTRISE, SCALED. default is "DEGREE".
    :param ps_power: double, used together with SCALED slope type, default is 0.664
    :param psz_factor: double, used together with SCALED slope type, default is 0.024
    :param remove_edge_effect: boolean, True or False, default is False
    :return: the output raster

    """

    template_dict = {
        "rasterFunction": "Slope",
        "rasterFunctionArguments": {
        },
        "variableName": "Raster"
    }

    slope_types = {
        "degree": 1,
        "percentrise": 2,
        "scaled": 3,
    }

    if slope_type.lower() not in slope_types:
        raise ValueError(
            "Invalid slope_type value. Note that slope_type should be one of 'DEGREE', 'PERCENTRISE', or 'SCALED'")

    if z_factor is not None:
        template_dict["rasterFunctionArguments"]["ZFactor"] = z_factor
    if slope_type is not None:
        template_dict["rasterFunctionArguments"]["SlopeType"] = slope_types[slope_type.lower()]
    if ps_power is not None:
        template_dict["rasterFunctionArguments"]["PSPower"] = ps_power
    if psz_factor is not None:
        template_dict["rasterFunctionArguments"]["PSZFactor"] = psz_factor
    if remove_edge_effect is not None:
        template_dict["rasterFunctionArguments"]["RemoveEdgeEffect"] = remove_edge_effect
    # if dem is not None:
    #     template_dict["rasterFunctionArguments"]["DEM"] = raster

    return Apply(dem, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Speckle(raster,
            filter_type="Lee",
            filter_size="3x3",
            noise_model="Multiplicative",
            noise_var=None,
            additive_noise_mean=None,
            multiplicative_noise_mean=1,
            nlooks=1,
            damp_factor=None,
            ):
    """
    The Speckle function filters the speckled radar dataset to smooth out the
    noise while retaining the edges or sharp features in the image. Four speckle
    reduction filtering algorithms are provided through this function. For more
    information including required and optional parameters for each filter and
    the default parameter values, see Speckle function
    https://pro.arcgis.com/en/pro-app/help/data/imagery/speckle-function.htm

    :param raster: input raster type
    :param filter_type: string, one of "Lee", "EnhancedLee" "Frost", "Kaun". Default is "Lee".
    :param filter_size: string, kernel size. One of "3x3", "5x5", "7x7", "9x9", "11x11". Default is "3x3".
    :param noise_model: string, For Lee filter only. One of "Multiplicative", "Additive", "AdditiveAndMultiplicative"
    :param noise_var: double, for Lee filter with noise_model "Additive" or "AdditiveAndMultiplicative"
    :param additive_noise_mean: double, for Lee filter witth noise_model "AdditiveAndMultiplicative" only
    :param multiplicative_noise_mean: double, For Lee filter with noise_model "Additive" or "AdditiveAndMultiplicative"
    :param nlooks: int, for Lee, EnhancedLee and Kuan Filters
    :param damp_factor: double, for EnhancedLee and Frost filters
    :return: output raster with function applied
    """
    filter_types = {
        "Lee": 0,
        "EnhancedLee": 1,
        "Frost": 2,
        "Kuan": 3
    }

    filter_sizes = {
        "3x3": 0,
        "5x5": 1,
        "7x7": 2,
        "9x9": 3,
        "11x11": 4
    }

    noise_models = {
        "Multiplicative": 0,
        "Additive": 1,
        "AdditiveAndMultiplicative": 2
    }

    template_dict = {
        "rasterFunction": "Speckle",
        "rasterFunctionArguments": {
        }
    }

    # value check
    if filter_type not in filter_types.keys():
        raise ValueError(
            "Invalid filter_type value. Note that filter_type should be 'Lee', 'EnhancedLee', 'Frost' or 'Kuan'")
    if filter_size not in filter_sizes.keys():
        raise ValueError(
            "Invalid filter_size value. Note that filter_size should be '3x3', '5x5', '7x7', '9x9' or '11x11'")
    if noise_model not in noise_models.keys():
        raise ValueError(
            "Invalid noise_model value. Note that noise_model should be 'Multiplicative', 'Additive', or 'AdditiveAndMultiplicative'")

    template_dict["rasterFunctionArguments"]['FilterType'] = filter_types[filter_type]
    template_dict["rasterFunctionArguments"]['FilterSize'] = filter_sizes[filter_size]
    template_dict["rasterFunctionArguments"]['NoiseModel'] = noise_models[noise_model]

    if noise_var is not None:
        template_dict["rasterFunctionArguments"]['NoiseVar'] = noise_var
    if additive_noise_mean is not None:
        template_dict["rasterFunctionArguments"]['AdditiveNoiseMean'] = additive_noise_mean
    if multiplicative_noise_mean is not None:
        template_dict["rasterFunctionArguments"]['MultiplicativeNoiseMean'] = multiplicative_noise_mean
    if nlooks is not None:
        template_dict["rasterFunctionArguments"]['NLooks'] = nlooks
    if damp_factor is not None:
        template_dict["rasterFunctionArguments"]['DampFactor'] = damp_factor

    return Apply(raster,
                 raster_function=template_dict['rasterFunction'],
                 raster_function_arguments=template_dict['rasterFunctionArguments'])


def SpectralConversion(raster, conversion_matrix):
    """
    The SpectralConversion function applies a matrix to a multi-band image to
    affect the spectral values of the output. In the matrix, different weights
    can be assigned to all the input bands to calculate each of the output
    bands. The column/row size of the matrix equals to the band count of input
    raster. For more information, see Spectral Conversion function
    https://pro.arcgis.com/en/pro-app/help/data/imagery/spectral-conversion-function.htm

    :param raster: the input raster
    :param conversion_matrix: array of double (A NxN length one-dimension matrix, where N=band count.)
    :return: the output raster with this function applied to it

    Example
    if conversion_matrix = [0.3,0.2,0.5,0,1], then, it means,
        Output band_1 = 0.3 * Input_band_1 + 0.2 * Input_band_2;
        Output band_2 = 0.5 * Input_band_1 + 0.1 * Input_band_2;
    """
    template_dict = {
        "rasterFunction": "SpectralConversion",
        "rasterFunctionArguments": {
            "ConversionMatrix": conversion_matrix
        }
    }

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def SRre(raster, nir_band_id=7, redEdge_band_id=6):
    """
    The Red-Edge Simple Ratio (SRre) is a vegetation index for estimating the
    amount of healthy and stressed vegetation. It is the ratio of light scattered
    in the NIR and red-edge bands, which reduces the effects of atmosphere and topography.

    Values are high for vegetation with high canopy closure and healthy vegetation,
    lower for high canopy closure and stressed vegetation, and low for soil, water,
    and nonvegetated features. The range of values is from 0 to about 30, where healthy
    vegetation generally falls between values of 1 to 10.

    SRre = NIR / RedEdge

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,7. Note that band ID uses one-based indexing.
    :param redEdge_band_id: the band ID of the redEdge band, e.g.,6. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [nir_band_id, redEdge_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")

    band_ids = str(nir_band_id) + " " + str(redEdge_band_id)
    return BandArithmetic(raster, band_ids, 13)


def StatisticsHistogram(raster, statistics=None, histograms=None):
    """"
    The function is used to define the statistics and histogram of a raster.
    It is normally used for control the default display of exported image.
    For more information, see Statistics and Histogram function,
    https://pro.arcgis.com/en/pro-app/help/data/imagery/statistics-and-histogram-function.htm

    :param raster: the input raster / imagery layer
    :param statistics: array of statistics objects. (Predefined statistics for each band)
    :param histograms: array of histogram objects. (Predefined histograms for each band)
    :return: Statistics and Histogram defined raster
    """
    template_dict = {
        "rasterFunction": "StatisticsHistogram",
        "rasterFunctionArguments": {}
    }

    if statistics is not None:
        template_dict["rasterFunctionArguments"]['Statistics'] = statistics
    if histograms is not None:
        template_dict["rasterFunctionArguments"]['Histograms'] = histograms

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Stretch(raster, stretch_type="None", min=0, max=255, num_stddev=2, statistics=None,
            dra=False, min_percent=0.25, max_percent=0.5, gamma=None, compute_gamma=False, sigmoid_strength_level=2):
    """
    The stretch function enhances an image through multiple stretch types.
    For more information, see https://pro.arcgis.com/en/pro-app/help/data/imagery/stretch-function.htm.
    Gamma stretch works with all stretch types.
    Min and Max can be used to define output minimum and maximum.
    DRA is used to get statistics from the extent in the export_image request.
    ComputeGamma will automatically calculate best gamma value to render exported image based on empirical model.
    Stretch type None does not require other parameters.
    Stretch type StdDev requires NumberOfStandardDeviations, Statistics, or DRA (true).
    Stretch type Histogram (Histogram Equalization) requires the source dataset to have histograms or additional DRA (true).
    Stretch type MinMax requires Statistics or DRA (true).
    Stretch type PercentClip requires MinPercent, MaxPercent, and DRA (true), or histograms from the source dataset.
    Stretch type Sigmoid does not require other parameters.
    Optionally, set the SigmoidStrengthLevel (1 to 6) to adjust the curvature of Sigmoid curve used in color stretch.

    The arguments for the stretch function are as follows:
    :param raster: input raster
    :param stretch_type: str, one of None, StdDev, Histogram, MinMax, PercentClip, Sigmoid
    :param min: double
    :param max: double
    :param num_stddev: double (e.g. 2)
    :param statistics: array of doubles, [[<min1>, <max1>, <mean1>, <standardDeviation1>], [<min2>, <max2>, <mean2>, <standardDeviation2>], ...]
    :param dra: True or False. Default is False. Statistics parameter is ignored when DRA is true
    :param min_percent: double (e.g. 0.25), applicable to PercentClip
    :param max_percent: double (e.g. 0.5), applicable to PercentClip
    :param gamma: array of doubles, e.g., [<gamma1>, <gamma2>, <gamma3>], one for each band
    :param compute_gamma: Trur or False. Default is False. applicable to any stretch type
    :param sigmoid_strength_level: int (1~6), applicable to Sigmoid
    :return: the output raster
    """
    str_types = {
        'none': 0,
        'stddev': 3,
        'histogram': 4,
        'minmax': 5,
        'percentclip': 6,
        'sigmoid': 9
    }
    
    if isinstance(stretch_type, str) and stretch_type.lower() in str_types:
        in_str_type = str_types[stretch_type.lower()]
    else:
        raise ValueError("Invalid stretch_type. stretch type should be one of None, StdDev, Histogram, MinMax, PercentClip")

    template_dict = {
        "rasterFunction": "Stretch",
        "rasterFunctionArguments": {},
    }

    if stretch_type is not None:
        template_dict["rasterFunctionArguments"]["StretchType"] = in_str_type
    if min is not None:
        template_dict["rasterFunctionArguments"]["Min"] = min
    if max is not None:
        template_dict["rasterFunctionArguments"]["Max"] = max
    if num_stddev is not None:
        template_dict["rasterFunctionArguments"]["NumberOfStandardDeviations"] = num_stddev
    if statistics is not None:
        template_dict["rasterFunctionArguments"]["Statistics"] = statistics
    if dra is not None:
        template_dict["rasterFunctionArguments"]["DRA"] = dra
    if min_percent is not None:
        template_dict["rasterFunctionArguments"]["MinPercent"] = min_percent
    if max_percent is not None:
        template_dict["rasterFunctionArguments"]["MaxPercent"] = max_percent
    if gamma is not None:
        template_dict["rasterFunctionArguments"]["Gamma"] = gamma
    if compute_gamma is not None:
        template_dict["rasterFunctionArguments"]["ComputeGamma"] = compute_gamma
    if sigmoid_strength_level is not None:
        template_dict["rasterFunctionArguments"]["SigmoidStrengthLevel"] = sigmoid_strength_level
    if compute_gamma is not None or gamma is not None:
        template_dict["rasterFunctionArguments"]["UseGamma"] = True

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Sultan(raster, band1_id=1, band3_id=3, band4_id=4, band5_id=5, band6_id=6):
    """
    Sultan's Formula (transform to 3 band 8 bit image)
        Band 1 = (Band5 / Band6) x 100
        Band 2 = (Band5 / Band1) x 100
        Band 3 = (Band3 / Band4) x (Band5 / Band4) x 100

    :param raster: the input raster
    :param band1_id(int): the band ID of band1, e.g., 1. Note that band ID uses one-based indexing.
    :param band3_id(int): the band ID of band3, e.g., 3. Note that band ID uses one-based indexing.
    :param band4_id(int): the band ID of band4, e.g., 4. Note that band ID uses one-based indexing.
    :param band5_id(int): the band ID of band5, e.g., 5. Note that band ID uses one-based indexing.
    :param band6_id(int): the band ID of band6, e.g., 6. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [band1_id, band3_id, band4_id, band5_id, band6_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")

    band_ids = str(band1_id) + ' ' + str(band3_id) + ' ' + str(band4_id) + ' ' + str(band5_id) + ' ' + str(band6_id)
    return BandArithmetic(raster, band_ids, 8)


def TasseledCap(raster):
    """"
    The function is designed to analyze and map vegetation and urban development
    changes detected by various satellite sensor systems. It is known as the
    Tasseled Cap transformation due to the shape of the graphical distribution
    of data. This function takes no arguments except a raster. The input for
    this function is the source raster of image service. There are no other
    parameters for this function because all the information is derived from
    the input's properties and key metadata (bands, data type, and sensor name).
    Only imagery from the Landsat MSS, Landsat TM, Landsat ETM+, IKONOS,
    QuickBird, WorldView-2 and RapidEye sensors are supported. Prior to applying
    this function, there should not be any functions that would alter the pixel
    values in the function chain, such as the Stretch, Apparent Reflectance or
    Pansharpening function. The only exception is for Landsat ETM+; when using
    Landsat ETM+, the Apparent Reflectance function must precede the Tasseled
    Cap function. For more information, see
    https://pro.arcgis.com/en/pro-app/help/data/imagery/tasseled-cap-function.htm

    :param raster: the input raster / imagery layer
    :return: the output raster with TasseledCap function applied to it
    """

    template_dict = {
        "rasterFunction": "TasseledCap",
        "rasterFunctionArguments": {}
    }

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def Threshold(raster):
    """
    The binary threshold function produces the binary image. It uses the Otsu method and assumes the input image to have
     a bi-modal histogram. The arguments for the threshold function are as follows:

    :param raster: input raster
    :return: the output raster

    """
    threshold_type = 1

    template_dict = {
        "rasterFunction": "Threshold",
        "rasterFunctionArguments": {
        },
        "variableName": "Raster"
    }

    if threshold_type is not None:
        template_dict["rasterFunctionArguments"]["ThresholdType"] = threshold_type

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def TransposeBits(raster, input_bit_positions=[12, 13], output_bit_positions=[0, 1], constant_fill_value=0,
                  fill_raster=None):
    """
    The transpose_bits function performs a bit operation. It extracts bit values from the source data and assigns them
    to new bits in the output data.The arguments for the transpose_bits function are as follows:

    Either providing a fill_value or a fill_raster to initialize pixel values of the output raster.
    Landsat 8 has a quality assessment band. The following are the example input and output bit positions to extract
    confidence levels by mapping them to 0-3:
    * Landsat 8 Water: {"input_bit_positions":[4,5],"output_bit_positions":[0,1]}
    * Landsat 8 Cloud Shadow: {"input_bit_positions":[6,7],"output_bit_positions":[0,1]}
    * Landsat 8 Vegetation: {"input_bit_positions":[8,9],"output_bit_positions":[0,1]}
    * Landsat 8 Snow/Ice: {"input_bit_positions":[10,11],"output_bit_positions":[0,1]}
    * Landsat 8 Cirrus: {"input_bit_positions":[12,13],"output_bit_positions":[0,1]}
    * Landsat 8 Cloud: {"input_bit_positions":[14,15],"output_bit_positions":[0,1]}
    * Landsat 8 Designated Fill: {"input_bit_positions":[0],"output_bit_positions":[0]}
    * Landsat 8 Dropped Frame: {"input_bit_positions":[1],"output_bit_positions":[0]}
    * Landsat 8 Terrain Occlusion: {"input_bit_positions":[2],"output_bit_positions":[0]}

    :param raster: input raster
    :param input_bit_positions: array of long, required
    :param output_bit_positions: array of long, required
    :param constant_fill_value: int, required
    :param fill_raster: optional, the fill raster
    :return: the output raster

    """

    template_dict = {
        "rasterFunction": "TransposeBits",
        "rasterFunctionArguments": {},
    }

    template_dict["rasterFunctionArguments"]["InputBitPositions"] = input_bit_positions
    template_dict["rasterFunctionArguments"]["OutputBitPositions"] = output_bit_positions

    if fill_raster is not None:
        template_dict["rasterFunctionArguments"]["FillRaster"] = fill_raster
        template_dict["rasterFunctionArguments"]["ConstantFillCheck"] = False
    else:
        if not isinstance(constant_fill_value, int):
            raise TypeError("constant_fill_value must be an integer")
        template_dict["rasterFunctionArguments"]["ConstantFillValue"] = constant_fill_value
    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def TSAVI(raster, nir_band_id=4, red_band_id=3, s=0.33, a=0.5, X=1.5):
    """
    Transformed Soil Adjusted Vegetation Index
    TSAVI = (s(NIR-s*Red-a))/(a*NIR+Red-a*s+X*(1+s^2))

    :param raster: the input raster / imagery layer
    :param nir_band_id: the band ID of the nir band, e.g.,4. Note that band ID uses one-based indexing.
    :param red_band_id: the band ID of the red band, e.g.,3. Note that band ID uses one-based indexing.
    :param s: the soil line slope, e.g.,0.33
    :param a: the soil line intercept, e.g.,0.5
    :param X: an adjustment factor that is set to minimize soil noise, e.g.,1.5
    :return: output raster
    """
    band_id_list = [nir_band_id, red_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")
    band_ids = str(nir_band_id) + ' ' + str(red_band_id) + ' ' + str(s) + ' ' + str(a) + ' ' + str(X)
    return BandArithmetic(raster, band_ids, 3)


def UnitConversion(raster, from_unit=None, to_unit=None):
    """
    The unit_conversion function performs unit conversions.The arguments for the unit_conversion function are as follows:
    from_unit and to_unit take the following str values:
    Speed Units: MetersPerSecond, KilometersPerHour, Knots, FeetPerSecond, MilesPerHour
    Temperature Units: Celsius,Fahrenheit,Kelvin
    Distance Units: str, one of Inches, Feet, Yards, Miles, NauticalMiles, Millimeters, Centimeters, Meters

    :param raster: input raster
    :param from_unit(string): value could be one of [MetersPerSecond, KilometersPerHour, Knots, FeetPerSecond, MilesPerHour,Celsius,Fahrenheit,Kelvin, Inches, Feet, Yards, Miles, NauticalMiles, Millimeters, Centimeters, Meters]
    :param to_unit(string): value could be one of [MetersPerSecond, KilometersPerHour, Knots, FeetPerSecond, MilesPerHour,Celsius,Fahrenheit,Kelvin, Inches, Feet, Yards, Miles, NauticalMiles, Millimeters, Centimeters, Meters]
    :return: the output raster

    """
    unit_types = {
        'inches': 1,
        'feet': 3,
        'yards': 4,
        'miles': 5,
        'nauticalmiles': 6,
        'millimeters': 7,
        'centimeters': 8,
        'meters': 9,
        'celsius': 200,
        'fahrenheit': 201,
        'kelvin': 202,
        'meterspersecond': 100,
        'kilometersperhour': 101,
        'knots': 102,
        'feetpersecond': 103,
        'milesperhour': 104
    }

    if isinstance(from_unit, str) and from_unit.lower() in unit_types.keys():
        from_unit = unit_types[from_unit.lower()]
    else:
        raise ValueError("Invalid from_unit value")

    if isinstance(to_unit, str) and to_unit.lower() in unit_types.keys():
        to_unit = unit_types[to_unit.lower()]
    else:
        raise ValueError("Invalid to_unit value")

    template_dict = {
        "rasterFunction": "UnitConversion",
        "rasterFunctionArguments": {}
    }

    if from_unit is not None:
        template_dict["rasterFunctionArguments"]["FromUnit"] = from_unit
    if to_unit is not None:
        template_dict["rasterFunctionArguments"]["ToUnit"] = to_unit

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def VARI(raster, red_band_id=3, green_band_id=2, blue_band_id=1):
    """
    Visible Atmospherically Resistant Index

    VARI = (Green - Red)/(Green + Red - Blue)

    :param raster: the input raster / imagery layer
    :param red_band_id:  the band ID of the red band, e.g.,3. Note that band ID uses one-based indexing.
    :param green_band_id: the band ID of the green band, e.g.,2. Note that band ID uses one-based indexing.
    :param blue_band_id: the band ID of the blue band, e.g.,1. Note that band ID uses one-based indexing.
    :return: output raster
    """
    band_id_list = [red_band_id, green_band_id, blue_band_id]
    for band_id in band_id_list:
        if not isinstance(band_id, int) or band_id <= 0:
            raise ValueError("band id should be positive integer value")

    band_ids = str(red_band_id) + ' ' + str(green_band_id) + ' ' + str(blue_band_id)
    return BandArithmetic(raster, band_ids, 9)


def VectorField(raster_u_mag, raster_v_dir, input_data_type='Vector-UV', angle_reference_system='Geographic',
                output_data_type='Vector-UV'):
    """
    The VectorField function is used to composite two single-band rasters (each raster represents U/V or Magnitude/Direction)
    into a two-band raster (each band represents U/V or Magnitude/Direction). Data combination type (U-V or Magnitude-Direction)
    can also be converted interchangeably with this function.
    For more information, see https://pro.arcgis.com/en/pro-app/help/data/imagery/vector-field-function.htm

    :param raster_u_mag: raster item representing 'U' or 'Magnitude' - imagery layers filtered by where clause, spatial and temporal filters
    :param raster_v_dir: raster item representing 'V' or 'Direction' - imagery layers filtered by where clause, spatial and temporal filters
    :param input_data_type: string, 'Vector-UV' or 'Vector-MagDir' per input used in 'raster_u_mag' and 'raster_v_dir'
    :param angle_reference_system: string, optional when 'input_data_type' is 'Vector-UV', one of "Geographic", "Arithmetic"
    :param output_data_type: string, 'Vector-UV' or 'Vector-MagDir'
    :return: the output raster with this function applied to it
    """
    angle_reference_system_types = {
        "Geographic": 0,
        "Arithmetic": 1
    }
    # value check
    if angle_reference_system not in angle_reference_system_types.keys():
        raise ValueError(
            "Invalid angle_reference_system value. Note that angle_reference_system should be either 'Geographic' or 'Arithmetic'")
    in_angle_reference_system = angle_reference_system_types[angle_reference_system]

    if input_data_type not in ["Vector-UV", "Vector-MagDir"]:
        raise ValueError(
            "Invalid input_data_type value. Note that input_data_type should be either 'Vector-UV' or 'Vector-MagDir'.")
    if output_data_type not in ["Vector-UV", "Vector-MagDir"]:
        raise ValueError(
            "Invalid output_data_type value. Note that output_data_type should be either 'Vector-UV' or 'Vector-MagDir'.")

    template_dict = {
        "rasterFunction": "VectorField",
        "rasterFunctionArguments": {}
    }

    template_dict["rasterFunctionArguments"]["AngleReferenceSystem"] = in_angle_reference_system
    template_dict["rasterFunctionArguments"]['InputDataType'] = input_data_type
    template_dict["rasterFunctionArguments"]['OutputDataType'] = output_data_type

    return Apply(in_raster=(raster_u_mag, raster_v_dir),
                 raster_function=template_dict['rasterFunction'],
                 raster_function_arguments=template_dict['rasterFunctionArguments'])


def VectorFieldRenderer(raster, is_uv_components=None, reference_system='Geographic',
                        mass_flow_angle_representation='From',
                        calculation_method="Vector Average", symbology_name="Single Arrow"):
    """
    The vector_field_renderer function symbolizes a U-V or Magnitude-Direction raster.The arguments for the vector_field_renderer function are as follows:

    :param raster: input raster
    :param is_uv_components: bool
    :param reference_system: int 1=Arithmetic, 2=Geographic
    :param mass_flow_angle_representation: int 0=from 1=to
    :param calculation_method: string, one of "Vector Average", 'Nearest neighbor', 'Bilinear', 'Cubic', 'Minimum', 'Maximum
    :param symbology_name: string, one of "Single Arrow" , "Wind Barb", "Ocean Current"
    :return: the output raster

    """
    mass_flow_angle_representations = {"from": 0, "to": 1}
    reference_systems = {"Arithmetic": 2, "Geographic": 1}

    if reference_system not in reference_systems.keys():
        raise ValueError(
            "Invalid reference_system. Note that reference_system should be either 'Geographic' or 'Arithmetic'")
    if mass_flow_angle_representation not in mass_flow_angle_representations.keys():
        raise ValueError(
            "Invalid mass_flow_angle_representation. Note that reference_system should be either'from' or 'to'")

    template_dict = {
        "rasterFunction": "VectorFieldRenderer",
        "rasterFunctionArguments": {}
    }

    if is_uv_components is not None:
        template_dict["rasterFunctionArguments"]["IsUVComponents"] = is_uv_components
    if reference_system is not None:
        template_dict["rasterFunctionArguments"]["ReferenceSystem"] = reference_systems[reference_system]
    if mass_flow_angle_representation is not None:
        template_dict["rasterFunctionArguments"]["MassFlowAngleRepresentation"] = \
            mass_flow_angle_representations[mass_flow_angle_representation]
    if calculation_method is not None:
        template_dict["rasterFunctionArguments"]["CalculationMethod"] = calculation_method
    if symbology_name is not None:
        template_dict["rasterFunctionArguments"]["SymbologyName"] = symbology_name

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def AspectSlope(raster, z_factor=1):
    """
    Creates a raster layer that simultaneously displays the aspect and slope of a surface.
    :param raster: input raster
    :param z_factor: A multiplication factor that converts the vertical (elevation) values to the linear units of the horizontal (x,y) coordinate system. "Use larger values to add vertical exaggeration.
    :return: the output raster
    """
    template_dict = {
        "rasterFunction": "AspectSlope",
        "rasterFunctionArguments": {}
    }

    template_dict['rasterFunctionArguments']['zf'] = z_factor
    template_dict['rasterFunctionArguments']['ClassName'] = "AspectSlope"

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])


def HeatIndex(temperature_raster, relative_humidity_raster, temperature_unit="Fahrenheit",
              heat_index_unit="Fahrenheit"):
    """
    Calculates apparent temperature based on ambient temperature and relative humidity. The apparent temperature is often described as how hot it feels to the human body.
    :param temperature_raster: the raster that represent temperature
    :param relative_humidity_raster: the raster that represent relative humidity
    :param temperature_unit: Available input units are Celsius,Fahrenheit or Kelvin
    :param heat_index_unit: Available input units are Celsius,Fahrenheit or Kelvin
    :return: the output raster
    """
    template_dict = {
        "rasterFunction": "HeatIndex",
        "rasterFunctionArguments": {}
    }
    if temperature_unit is not None:
        template_dict["rasterFunctionArguments"]["units"] = temperature_unit
    if heat_index_unit is not None:
        template_dict["rasterFunctionArguments"]["outUnits"] = heat_index_unit
    template_dict["rasterFunctionArguments"]["MatchVariable"] = False

    return Apply({"temperature": temperature_raster, "rh": relative_humidity_raster}, template_dict['rasterFunction'],
                 template_dict['rasterFunctionArguments'])


def WindChill(temperature_raster, wind_speed_raster, temperature_units="Fahrenheit", wind_speed_units="mph",
              wind_chill_units="Fahrenheit"):
    """
    The Wind Chill function is useful for identifying dangerous winter conditions that, depending on exposure times to
    the elements, can result in frostbite or even hypothermia. Wind chill is a way to measure how cold an individual
    feels when wind is taken into account with already cold temperatures. The faster the wind speed, the more quickly the body will lose heat and the colder they will feel.

    :param temperature_raster:A single-band raster where pixel values represent ambient air temperature.
    :param wind_speed_raster:A single-band raster where pixel values represent wind speed.
    :param temperature_units:The unit of measurement associated with the input temperature raster. Available input units are Celsius, Fahrenheit, and Kelvin.
    :param wind_speed_units: Defines the unit of measurement for the wind-speed raster. Available input units are mph, km/h, ft/s and kn. Each represents

                        Miles Per Hour (mph)
                        Kilometers Per Hour (km/h)
                        Meters Per Second (m/s)
                        Feet Per Second (ft/s)
                        Knots (kn)
    :param wind_chill_units:The unit of measurement associated with the output raster. Available output units are Celsius, Fahrenheit, and Kelvin.
    :return: the output raster
    """
    template_dict = {
        "rasterFunction": "Windchill",
        "rasterFunctionArguments": {}
    }

    if temperature_units is not None:
        template_dict["rasterFunctionArguments"]["tunits"] = temperature_units
    if wind_speed_units is not None:
        template_dict["rasterFunctionArguments"]["wunits"] = wind_speed_units
    if wind_chill_units is not None:
        template_dict["rasterFunctionArguments"]["ounits"] = wind_chill_units
    template_dict["rasterFunctionArguments"]["MatchVariable"] = False
    return Apply({"temperature": temperature_raster, "ws": wind_speed_raster}, template_dict['rasterFunction'],
                 template_dict['rasterFunctionArguments'])


def Contour(raster, adaptive_smoothing=2.5, contour_type="contour lines", z_base=0, number_of_contours=0,
            contour_interval=100, nth_contour_line_in_bold=5, z_factor=1):
    """
    Creates contour lines.

    :param raster: the raster from which the contour is created
    :param adaptive_smoothing(double): adaptive smooting value, e.g., 2.5
    :param contour_type(string): contour type. Available values could be "contour lines", "contour fill" or "smooth surface only"
    :param z_base(double): the z-base value, e.g., 0
    :param number_of_contours(int): the number of contours, e.g., 0
    :param contour_interval(double): the contour interval, e.g., 100
    :param nth_contour_line_in_bold(int): the nth contour line that would be rendered in bold, e.g., 5
    :param z_factor(double): the z-factor, e.g., 1
    :return:
    """
    template_dict = {
        "rasterFunction": "Contour",
        "rasterFunctionArguments": {}
    }

    contour_types = {
        'contour lines': 0,
        'contour fill': 1,
        'smooth surface only': 2,
    }

    if adaptive_smoothing is not None:
        template_dict['rasterFunctionArguments']['SigmaGaussian'] = adaptive_smoothing
    if contour_type is not None:
        template_dict['rasterFunctionArguments']['ContourType'] = contour_types[contour_type.lower()]
    if z_base is not None:
        template_dict['rasterFunctionArguments']['ZBase'] = z_base
    if number_of_contours is not None:
        template_dict['rasterFunctionArguments']['NumberOfContours'] = number_of_contours
    if contour_interval is not None:
        template_dict['rasterFunctionArguments']['ContourInterval'] = contour_interval
    if nth_contour_line_in_bold is not None:
        template_dict['rasterFunctionArguments']['NthContourLineInBold'] = nth_contour_line_in_bold
    if z_factor is not None:
        template_dict['rasterFunctionArguments']['ZFactor'] = z_factor

    return Apply(raster, template_dict['rasterFunction'], template_dict['rasterFunctionArguments'])
