# -*- coding: utf-8 -*-
#COPYRIGHT 2019 ESRI
#
#TRADE SECRETS: ESRI PROPRIETARY AND CONFIDENTIAL
#Unpublished material - all rights reserved under the
#Copyright Laws of the United States.
#
#For additional information, contact:
#Environmental Systems Research Institute, Inc.
#Attn: Contracts Dept
#380 New York Street
#Redlands, California, USA 92373
#
#email: contracts@esri.com
r"""The Maritime toolbox contains tools that manage S-100 and bathymetric
data."""
__all__ = ['ConvertS57ToS101', 'ExportS101Cell', 'ImportS100Cell',
           'ImportS100FeatureCatalogue', 'ReducePointDensity',
           'SmoothBathymetricTIN']
__alias__ = 'maritime'
from arcpy.geoprocessing._base import gptooldoc, gp, gp_fixargs
from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject

# Tools
@gptooldoc('ReducePointDensity_maritime', None)
def ReducePointDensity(in_features=None, out_feature_class=None, depth_field=None, depth_direction=None, depth_bias=None, radius_unit=None, start_thinning_radius=None, end_thinning_radius=None, start_depth=None, end_depth=None):
    """ReducePointDensity_maritime(in_features, out_feature_class, depth_field, depth_direction, depth_bias, radius_unit, start_thinning_radius, {end_thinning_radius}, {start_depth}, {end_depth})

        Thins points from a point or multipoint feature class.Using an input
        feature class with either a numeric depth field or
        z-enabled geometry, the tool generates an output feature class with a
        set of points that have been thinned to a user-specified radius in a
        shallow or deep-biased manner.

     INPUTS:
      in_features (Feature Layer):
          The input point or multipoint features.
      depth_field (Field):
          The field where the depth is stored. It is either a numeric field or
          the shape field specified in Input Features.  For multipoint features,
          this must be the shape field.
      depth_direction (String):
          Specifies how the depth will be captured in the input TIN.

          * POSITIVE_UP-Depth values are positive above the surface and negative
          below the surface. This is default.

          * POSITIVE_DOWN-Depth values are positive below the surface and
          negative above the surface.
      depth_bias (String):
          Specifies the bias that will be used to select the depths to be
          retained.

          * SHALLOW_BIASED-Shallow bias will be used for depth. This is default.

          * DEEP_BIASED-Deep bias will be used for depth.
      radius_unit (String):
          Specifies the unit of measure used by the Start Thinning Radius and
          End Thinning Radius parameters.

          * KILOMETERS-The depth value in kilometers.

          * METERS-The depth value in meters. This is default.

          * DECIMETERS-The depth value in decimeters.

          * CENTIMETERS-The depth value in centimeters.

          * MILLIMETERS-The depth value in millimeters.

          * NAUTICAL_MILES-The depth value in nautical miles.

          * MILES-The depth value in miles.

          * YARDS-The depth value in yards.

          * FEET-The depth value in feet.

          * INCHES-The depth value in inches.

          * DECIMAL_DEGREES-The depth value in decimal degrees.

          * POINTS-The depth value in points.
      start_thinning_radius (Double):
          The beginning radius that will be used to remove or thin points
          relative to each other.
      end_thinning_radius {Double}:
          The end radius that will be used to remove or thin points relative to
          each other. The thinning radius will dynamically change as the
          algorithm progresses through the range of depth values.
      start_depth {Double}:
          The depth that will be used to begin the thinning algorithm. Depth
          values that appear before this depth based on Depth Direction will be
          ignored.
      end_depth {Double}:
          The depth that will be used to end the thinning algorithm. Depth
          values that appear after this depth based on Depth Direction will be
          ignored.

     OUTPUTS:
      out_feature_class (Feature Class):
          The output feature class."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ReducePointDensity_maritime(*gp_fixargs((in_features, out_feature_class, depth_field, depth_direction, depth_bias, radius_unit, start_thinning_radius, end_thinning_radius, start_depth, end_depth), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('SmoothBathymetricTIN_maritime', None)
def SmoothBathymetricTIN(in_tin=None, out_tin=None, depth_direction=None, smoothing_iterations=None):
    """SmoothBathymetricTIN_maritime(in_tin, out_tin, depth_direction, smoothing_iterations)

        Smooths a triangulated irregular network (TIN) dataset in a manner
        that strictly preserves a shallow bias. The output is used to generate
        depth contours used for navigation, so
        the output TIN nodes will never report a value deeper than the input.

     INPUTS:
      in_tin (TIN Layer):
          The input TIN that will be smoothed with a shallow bias.
      depth_direction (String):
          Specifies how the depth will be captured in the input TIN.

          * POSITIVE_UP-The depth will be captured in the input TIN. This is
          default.

          * POSITIVE_DOWN-The downward depth will be captured in the input TIN.
      smoothing_iterations (Long):
          The number of smoothing passes that will be performed over the TIN.

     OUTPUTS:
      out_tin (TIN):
          The output smoothed TIN."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.SmoothBathymetricTIN_maritime(*gp_fixargs((in_tin, out_tin, depth_direction, smoothing_iterations), True)))
        return retval
    except Exception as e:
        raise e


# S-100 toolset
@gptooldoc('ConvertS57ToS101_maritime', None)
def ConvertS57ToS101(in_feature_catalogue=None, in_config_file=None, input_s57=None, out_location=None):
    """ConvertS57ToS101_maritime(in_feature_catalogue, in_config_file, input_s57, out_location)

        Converts the S-57 vector product for storing nautical charting data to
        the new vector S-101 format.

     INPUTS:
      in_feature_catalogue (File):
          The S-101 feature catalogue (XML document) from the International
          Hydrographic Organization (IHO) containing the schema of the features,
          attributes, and relationships used for encoding the hydrographic data
          in the S-101 cells.
      in_config_file (File):
          The input XML file that can be used to customize some aspects of the
          conversion process.
      input_s57 (File / Folder):
          The input S-57 file with a .000 extension or a CATALOG.031 file that
          references a collection of S-57 files.
      out_location (Folder):
          The directory where the converted cell will be written."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ConvertS57ToS101_maritime(*gp_fixargs((in_feature_catalogue, in_config_file, input_s57, out_location), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('ExportS101Cell_maritime', None)
def ExportS101Cell(in_feature_catalogue=None, in_s101_workspace=None, product=None, export_type=None, output_location=None):
    """ExportS101Cell_maritime(in_feature_catalogue, in_s101_workspace, product, export_type, output_location)

        Exports S-101 hydrographic data from a geodatabase to an S-101 file.

     INPUTS:
      in_feature_catalogue (File):
          An  S-100 feature catalogue document that describes the content of a
          data product and specification.
      in_s101_workspace (Workspace):
          The workspace that contains the product.
      product (String):
          The name of the product to export.This information must exist in the
          ProductCoverage feature class and
          ProductDefinition table before using this tool.
      export_type (String):
          The type of file created during export.

          * NEW_EDITION-A new edition of a dataset, including new information
          that has not been previously distributed by updates. This is the
          default.
      output_location (Folder):
          The location where the export package will be written."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ExportS101Cell_maritime(*gp_fixargs((in_feature_catalogue, in_s101_workspace, product, export_type, output_location), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('ImportS100Cell_maritime', None)
def ImportS100Cell(in_feature_catalogue=None, in_base_cell=None, target_workspace=None, in_update_cells=None):
    """ImportS100Cell_maritime(in_feature_catalogue, in_base_cell, target_workspace, {in_update_cells;in_update_cells...})

        Imports S-100 hydrographic data into a geodatabase created from a
        related S-100 feature catalogue.

     INPUTS:
      in_feature_catalogue (File):
          An  S-100 feature catalogue document that describes the content of a
          data product and specification.
      in_base_cell (File):
          The base file in the S-100 format.
      target_workspace (Workspace):
          The geodatabase to which all output data will be written.
      in_update_cells {File}:
          An update to a base file in the S-100 format."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ImportS100Cell_maritime(*gp_fixargs((in_feature_catalogue, in_base_cell, target_workspace, in_update_cells), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('ImportS100FeatureCatalogue_maritime', None)
def ImportS100FeatureCatalogue(in_feature_catalogue=None, target_workspace=None, admin_connection=None):
    """ImportS100FeatureCatalogue_maritime(in_feature_catalogue, target_workspace, {admin_connection})

        Imports the contents of an S-100 feature catalogue into an existing
        geodatabase. A feature catalogue is an XML document that describes the
        content of a data product.Examples of S-100 data products include the
        following:

        * S-101 Electronic Navigation Chart

        * S-111 Surface Currents

     INPUTS:
      in_feature_catalogue (File):
          An S-100 feature catalogue document that describes the content of a
          data product and specification.
      target_workspace (Workspace):
          The geodatabase to which output data will be written.
      admin_connection {File}:
          Optional enterprise geodatabase administrator connection file that
          should be used when importing into an enterprise geodatabase, for
          example, the SDE user."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ImportS100FeatureCatalogue_maritime(*gp_fixargs((in_feature_catalogue, target_workspace, admin_connection), True)))
        return retval
    except Exception as e:
        raise e


# End of generated toolbox code
del gptooldoc, gp, gp_fixargs, convertArcObjectToPythonObject