# -*- coding: utf-8 -*-
#COPYRIGHT 2019 ESRI
#
#TRADE SECRETS: ESRI PROPRIETARY AND CONFIDENTIAL
#Unpublished material - all rights reserved under the
#Copyright Laws of the United States.
#
#For additional information, contact:
#Environmental Systems Research Institute, Inc.
#Attn: Contracts Dept
#380 New York Street
#Redlands, California, USA 92373
#
#email: contracts@esri.com
r"""The Trace Network toolbox contains tools to create, configure, and
work with trace networks."""
__all__ = ['BuildTraceNetwork', 'ConvertGeometricNetworkToTraceNetwork',
           'CreateTraceNetwork', 'SetFlowDirection', 'Trace']
__alias__ = 'tn'
from arcpy.geoprocessing._base import gptooldoc, gp, gp_fixargs
from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject

# Tools
@gptooldoc('BuildTraceNetwork_tn', None)
def BuildTraceNetwork(in_trace_network=None, extent=None):
    """BuildTraceNetwork_tn(in_trace_network, {extent})

        Builds the network topology between feature classes in a trace
        network.

     INPUTS:
      in_trace_network (Trace Network / Trace Network Layer):
          The trace network for which the network topology will be built.
      extent {Extent}:
          Extent"""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.BuildTraceNetwork_tn(*gp_fixargs((in_trace_network, extent), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('SetFlowDirection_tn', None)
def SetFlowDirection(input_trace_network=None, in_edges=None, flow_direction=None):
    """SetFlowDirection_tn(input_trace_network, in_edges;in_edges..., flow_direction)

        Set the flow direction of features in a trace network.

     INPUTS:
      input_trace_network (Trace Network / Trace Network Layer):
          The trace network that contains the edges class to set the flow
          direction for.
      in_edges (Feature Layer):
          The polyline feature classes or feature layers that participates in
          the trace network to set the flow direction for. If there is a
          selection on the feature layer, the flow selection will be set for
          only selected features.
      flow_direction (String):
          Set the flow direction for the edges.

          * WITH_DIGITIZED_DIRECTION-Establish flow direction along the
          digitized direction of edges.

          * AGAINST_DIGITIZED_DIRECTION-Establish flow direction against the
          digitized direction of edges.

          * INDETERMINATE-Establish flow direction as indeterminate."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.SetFlowDirection_tn(*gp_fixargs((input_trace_network, in_edges, flow_direction), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('Trace_tn', None)
def Trace(in_trace_network=None, trace_type=None, starting_points=None, barriers=None, allow_indeterminate_flow=None, path_direction=None, path_network_weight_name=None, condition_barriers=None, function_barriers=None, traversability_scope=None, functions=None, output_conditions=None):
    """Trace_tn(in_trace_network, trace_type, {starting_points}, {barriers}, {allow_indeterminate_flow}, {path_direction}, {path_network_weight_name}, {condition_barriers;condition_barriers...}, {function_barriers;function_barriers...}, {traversability_scope}, {functions;functions...}, {output_conditions;output_conditions...})

        Returns selected features based on connectivity or traversability from
        the specified starting points.Trace network tracing capabilities allow
        you to analyze the paths in
        your network. This tool runs a trace that returns features based on
        connectivity or traversability from the specified starting points.
        Traversability refers to paths established by connected or associated
        features that also meet configuration requirements. Results from a
        trace operation are returned as a selection set of the
        appropriate network features. The resulting selection set of a trace
        can then be propagated to another map or used as input to another tool
        or trace.

     INPUTS:
      in_trace_network (Trace Network / Trace Network Layer):
          The trace network on which the trace will be run.
      trace_type (String):
          The type of trace to execute.

          * CONNECTED-A connected trace that begins at one or more starting
          points and spans outward along connected features.

          * UPSTREAM-An upstream trace that discovers features upstream from a
          location in the network. This type of trace requires the flow
          direction to be set.

          * DOWNSTREAM-A downstream trace that discovers features downstream
          from a location in the network. This type of trace requires the flow
          direction to be set.

          * PATH-This type of trace finds a path between two starting points in
          the network, regardless of flow direction. The cost of traversing the
          path is determined based on the network weight set for the
          path_network_weight_name value.
      starting_points {Feature Layer}:
          A point feature class or feature layer that contains one or more
          features that represents the starting points of the trace. This
          feature class is generated in the default geodatabase of the project
          when you create starting points using the Starting Points tool on the
          Trace Locations pane and will be used by default.
      barriers {Feature Layer}:
          A point feature class or feature layer that contains features that
          represents the barriers of the trace. This feature class is generated
          in the default geodatabase of the project when you create barriers
          using the Barriers tool on the Trace Locations pane and will be used
          by default.
      allow_indeterminate_flow {Boolean}:
          Indicates whether features that have indeterminate or uninitialized
          flow should be traced. This parameter is only honored when running and
          upstream or downstream trace type.

          * TRACE_INDETERMINATE_FLOW-Include features that have indeterminate or
          uninitialized flow direction in the trace.

          * IGNORE_INDETERMINATE_FLOW-Do not trace features that have
          indeterminate or uninitialized flow direction. This is the default.
      path_direction {String}:
          The direction of the path trace. The cost of traversing the path is
          determined based on the value set for the path_network_weight_name.
          This parameter is only honored when running a PATH trace type.

          * NO_DIRECTION-Find the path between the two starting points
          regardless of the direction of flow. This is the default.

          * PATH_UPSTREAM-Find the downstream path between the two starting
          points.

          * PATH_DOWNSTREAM-Find the upstream path between the two starting
          points.
      path_network_weight_name {String}:
          The network weight used to calculate the path. When running a path
          trace type, the shortest path is calculated using a numeric network
          weight such as shape length. Cost and distance-based paths can both be
          achieved. The default value is a path with the least number of
          features, which is equivalent to the geometric network. This parameter
          is required when running a path trace.
      condition_barriers {Value Table}:
          Sets a traversability barrier condition on features based on a
          comparison to a network weight. A condition barrier uses a network
          attribute, an operator and a type, and an attribute value. For
          example, stop a trace when a feature has the attribute Device Status
          equal to the specific value of Open. When a feature meets this
          condition, the trace stops. If you're using more than one attribute,
          you can use the Combine Using parameter to define an And or an Or
          condition.Condition barrier components are as follows:

          * Name-Choose to filter by any network weight defined in the system.

          * Operator-Choose from a number of different operators.

          * Type-Choose a specific value or network weight from the value that
          is specified in the name parameter.

          * Value-Set a specific value of the input attribute type that would
          cause termination based on the operator value.

          * Combine Using-Set this value if you have multiple attributes to add.
          You can combine them using an And or an Or condition.
          The condition barriers operator values are as follows:

          * IS_EQUAL_TO-The attribute is equal to the value.

          * DOES_NOT_EQUAL-The attribute is not equal to the value.

          * IS_GREATER_THAN-The attribute is greater than the value.

          * IS_GREATER_THAN_OR_EQUAL_TO-The attribute is greater than or equal
          to the value.

          * IS_LESS_THAN-The attribute is less than the value.

          * IS_LESS_THAN_OR_EQUAL_TO-The attribute is less than or equal to the
          value.

          * INCLUDES_THE_VALUES-A bitwise AND operation where all bits in the
          value are present in the attribute (bitwise AND == value).

          * DOES NOT INCLUDE_THE_VALUES-A bitwise AND operation where not all of
          the bits in the value are present in the attribute (bitwise AND !=
          value).

          * INCLUDES_ANY-A bitwise AND operation where at least one bit in the
          value is present in the attribute (bitwise AND == True).

          * DOES_NOT_INLCUDE_ANY-A bitwise AND operation where none of the bits
          in the value are present in the attribute (bitwise AND == False).
          Possible values for the Type options are as follows:

          * SPECIFIC_VALUE-Filter by a specific value.

          * NETWORK_ATTRIBUTE-Filter by a network weight.
          The Combine Using values are as follows:

          * AND-Combine the condition barriers.

          * OR-Use if either condition barrier is met.
      function_barriers {Value Table}:
          Sets a traversability barrier on features based on a function.
          Function barriers can be used to do such things as restrict how far
          the trace travels from the starting point, or set a maximum value to
          stop a trace. For example, the length of each line traveled is added
          to the total distance traveled so far. When the total length traveled
          reaches the value specified, the trace stops.Function barrier
          components are as follows:

          * Function-Choose from a number of different calculation functions.

          * Attribute-Choose to filter by any network weight defined in the
          system.

          * Operator-Choose from a number of different operators.

          * Value-Set a specific value of the input attribute type that, if
          discovered, will cause the termination.

          * Use Local Values-Calculate values in each direction as opposed to an
          overall global value, for example, for a function barrier that is
          calculating the sum of Shape length where the trace terminates if the
          value is greater than or equal to 4. In the global case, after you
          have traversed two edges with a value of 2, you have already reached a
          shape length sum of 4, so the trace stops. If local values are used,
          the local values along each path change, so the trace goes farther.
          Possible values for the function barrier function options are as
          follows:

          * AVERAGE-The average of the input values.

          * COUNT-The number of features.

          * MAX-The maximum of the input values.

          * MIN-The minimum of the input values.

          * ADD-Add the values.

          * SUBTRACT-Subtract the values.
          The function barrier operator value options are as follows:

          * IS_EQUAL_TO-The attribute is equal to the value.

          * DOES_NOT_EQUAL-The attribute is not equal to the value.

          * IS_GREATER_THAN-The attribute is greater than the value.

          * IS_GREATER_THAN_OR_EQUAL_TO-The attribute is greater than or equal
          to the value.

          * IS_LESS_THAN-The attribute is less than the value.

          * IS_LESS_THAN_OR_EQUAL_TO-The attribute is less than or equal to the
          value.

          * INCLUDES_THE_VALUES-A bitwise AND operation where all bits in the
          value are present in the attribute (bitwise AND == value).

          * DOES_NOT_INCLUDE_THE_VALUES-A bitwise AND operation where not all of
          the bits in the value are present in the attribute (bitwise AND !=
          value).

          * INCLUDES_ANY-A bitwise AND operation where at least one bit in the
          value is present in the attribute (bitwise AND == True).

          * DOES_NOT_INCLUDE_ANY-A bitwise AND operation where none of the bits
          in the value are present in the attribute (bitwise AND == False).
          The function barrier Use Local Values options are as follows:

          * TRUE-Use local values.

          * FALSE-Use global values. This is the default.
      traversability_scope {String}:
          The type of traversability to enforce. Traversability scope dictates
          whether traversability is enforced at junctions, edges, or both. For
          example, if a condition barrier is defined to stop the trace if Device
          Status is equal to Open and traversability scope is set to edges only,
          the trace would not stop even if the trace encounters an open device,
          because Device Status is only applicable for junctions. In other
          words, this parameter indicates to the trace whether or not to ignore
          junctions, edges, or both.

          * BOTH_JUNCTIONS_AND_EDGES-Apply traversability to both junctions and
          edges. This is the default.

          * JUNCTIONS_ONLY-Apply traversability to only junctions.

          * EDGES_ONLY-Apply traversability to only edges.
      functions {Value Table}:
          Applies a calculation function to the trace results.Functions
          components are as follows:

          * Function-Choose from a number of calculation functions.

          * Attribute-Filter by any network weight defined in the system.

          * Filter Name-Filter the function results by attribute name.

          * Filter Operator-Choose from a number of operators.

          * Filter Type-Choose from a number of filter types.

          * Filter Value-Provide a specific value for the input filter
          attribute.
          The functions Function value options are as follows:

          * AVERAGE-The average of the input values.

          * COUNT-The number of features.

          * MAX-The maximum of the input values.

          * MIN-The minimum of the input values.

          * ADD-The sum of the values.

          * SUBTRACT-The difference between the values.
          For example, there is a starting point feature with a value of 20. The
          next feature has a value of 30. If you are using the MINIMUM function,
          the result is 20. MAXIMUM is 30, ADD is 50, AVERAGE is 25, COUNT is 2,
          and SUBTRACT is -10.The Filter Operator value options are:

          * IS_EQUAL_TO-The attribute is equal to the value.

          * DOES_NOT_EQUAL-The attribute is not equal to the value.

          * IS_GREATER_THAN-The attribute is greater than the value.

          * IS_GREATER_THAN_OR_EQUAL_TO-The attribute is greater than or equal
          to the value.

          * IS_LESS_THAN-The attribute is less than the value.

          * IS_LESS_THAN_OR_EQUAL_TO-The attribute is less than or equal to the
          value.

          * INCLUDES_THE_VALUES-A bitwise AND operation where all bits in the
          value are present in the attribute (bitwise AND == value).

          * DOES_NOT_INCLUDE_THE_VALUES-A bitwise AND operation where not all of
          the bits in the value are present in the attribute (bitwise AND !=
          value).

          * INCLUDES_ANY-A bitwise AND operation where at least one bit in the
          value is present in the attribute (bitwise AND == True).

          * DOES_NOT_INCLUDE_ANY-A bitwise AND operation where none of the bits
          in the value are present in the attribute (bitwise AND == False).
          The functions Filter Type value options are as follows:

          * SPECIFIC_VALUE-Filter by a specific value.

          * NETWORK_ATTRIBUTE-Filter by a network weight.
      output_conditions {Value Table}:
          Filter the output subtypes to be included in the results. For
          example, only return overheard transformers."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.Trace_tn(*gp_fixargs((in_trace_network, trace_type, starting_points, barriers, allow_indeterminate_flow, path_direction, path_network_weight_name, condition_barriers, function_barriers, traversability_scope, functions, output_conditions), True)))
        return retval
    except Exception as e:
        raise e


# Administration toolset
@gptooldoc('ConvertGeometricNetworkToTraceNetwork_tn', None)
def ConvertGeometricNetworkToTraceNetwork(in_geometric_network=None, out_trace_network_name=None):
    """ConvertGeometricNetworkToTraceNetwork_tn(in_geometric_network, out_trace_network_name)

        Convert a geometric network to a trace network.

     INPUTS:
      in_geometric_network (Geometric Network):
          The geometric network that will be converted to a trace network.
          Converting a geometric network to a trace network will drop the
          geometric network and create a trace network in-place. There is no way
          to undo this change. Ensure you have a backup of your data before
          proceeding.
      out_trace_network_name (String):
          The name of the output trace network."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ConvertGeometricNetworkToTraceNetwork_tn(*gp_fixargs((in_geometric_network, out_trace_network_name), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('CreateTraceNetwork_tn', None)
def CreateTraceNetwork(in_feature_dataset=None, in_trace_network_name=None, input_junctions=None, input_edges=None):
    """CreateTraceNetwork_tn(in_feature_dataset, in_trace_network_name, {input_junctions;input_junctions...}, {input_edges;input_edges...})

        Creates a trace network in a file geodatabase feature dataset. A
        trace network dataset is created based on the input point and line
        feature classes that will participate in the trace network.

     INPUTS:
      in_feature_dataset (Feature Dataset):
          The feature dataset that will contain the trace network.
      in_trace_network_name (String):
          The name of the trace network that will be created.
      input_junctions {String}:
          The names of the point feature classes within the feature dataset to
          include in the trace network.
      input_edges {Value Table}:
          The names of the line feature classes within the feature dataset to
          include in the trace network. For each line feature class, you must
          also specify an associated connectivity policy.

          * SIMPLE_EDGE-Allows resources to flow from one end of the edge and
          out the other end.

          * COMPLEX_EDGE-Allows resources to be siphoned off along the length of
          the edge."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.CreateTraceNetwork_tn(*gp_fixargs((in_feature_dataset, in_trace_network_name, input_junctions, input_edges), True)))
        return retval
    except Exception as e:
        raise e


# End of generated toolbox code
del gptooldoc, gp, gp_fixargs, convertArcObjectToPythonObject